from .start_auto_reloader_ import *
from .stop_auto_reloader_ import *
from .update_auto_reloader_ import *
from .warn_auto_reloader_availability_ import *


__all__ = (
    *start_auto_reloader_.__all__,
    *stop_auto_reloader_.__all__,
    *update_auto_reloader_.__all__,
    *warn_auto_reloader_availability_.__all__,
)



from .activity import *
from .activity_assets import *
from .activity_field_base import *
from .activity_metadata import *
from .activity_party import *
from .activity_timestamps import *


__all__ = (
    *activity.__all__,
    *activity_assets.__all__,
    *activity_field_base.__all__,
    *activity_metadata.__all__,
    *activity_party.__all__,
    *activity_timestamps.__all__,
)

# Construct deprecations

from scarletio import modulize

from ...utils.module_deprecation import deprecated_import


@deprecated_import
@modulize
class ACTIVITY_TYPES:
    """
    Deprecated. Please use ``ActivityType`` instead.
    """
    game = 0
    stream = 1
    spotify = 2
    watching = 3
    custom = 4
    competing = 5


deprecated_import(Activity, 'ActivityRich')



__all__ = ()

from .preinstanced import AutoModerationActionType


KEYWORD_TO_ACTION_TYPE = {
    'channel_id': AutoModerationActionType.send_alert_message,
    'duration': AutoModerationActionType.timeout,
    'custom_message': AutoModerationActionType.block_message,
}


def guess_action_type_from_keyword_parameters(action_type, keyword_parameters):
    """
    Guesses the action's type from the given fields.
    
    Parameters
    ----------
    action_type : ``AutoModerationActionType``
        Already detected action type.
    
    keyword_parameters : `dict` of (`str`, `object`) items
        Keyword parameters passed.
    
    Returns
    -------
    action_type : ``AutoModerationActionType``
        The detected type.
    
    Raises
    ------
    TypeError
        - Extra keyword parameter.
        - Multiple actions received / detected.
    """
    for key in keyword_parameters.keys():
        try:
            key_action_type = KEYWORD_TO_ACTION_TYPE[key]
        except KeyError:
            raise TypeError(
                f'Extra keyword parameter: {key} = {keyword_parameters[key]!r}.'
            )
        
        if action_type is AutoModerationActionType.none:
            action_type = key_action_type
            continue
        
        if action_type is not key_action_type:
            raise TypeError(
                f'Multiple action types received / detected: {action_type!r}; {key_action_type!r}; '
                f'keyword parameters = {keyword_parameters!r}.'
            )
    
    return action_type



__all__ = ('EULA', )

from ...bases import DiscordEntity
from ...core import EULAS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_content, parse_id, parse_name, put_content_into, put_id_into, put_name_into, validate_content, validate_id,
    validate_name
)


PRECREATE_FIELDS = {
    'content': ('content', validate_content),
    'name': ('name', validate_name),
}


class EULA(DiscordEntity, immortal = True):
    """
    Represents a Discord end-user license agreement
    
    Attributes
    ----------
    content : `None`, `str`
        The eula's content.
    id : `int`
        The unique identifier number of the eula.
    name : `str`
        The eula's name.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = ('content', 'name')
    
    
    def __new__(cls, *, content = ..., name = ...):
        """
        Creates a new partial application eula.
        
        Parameters
        ----------
        content : `None`, `str`, Optional (Keyword only)
            The eula's content.
        
        name : `str`, Optional (Keyword only)
            The eula's name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # content
        if content is ...:
            content = None
        else:
            content = validate_content(content)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        self = object.__new__(cls)
        self.content = content
        self.id = 0
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new eula instance from the given data.
        
        If the eula already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Eula data.
        """
        eula_id = parse_id(data)
        
        try:
            self = EULAS[eula_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = eula_id
            EULAS[eula_id] = self
        
        self._update_attributes(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the eula into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        if include_internals:
            put_id_into(self.id, data, defaults)
        
        put_content_into(self.content, data, defaults)
        put_name_into(self.name, data, defaults)
        
        return data
    
    
    def _update_attributes(self, data):
        """
        Updates the eula with the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        self.content = parse_content(data)
        self.name = parse_name(data)
    
    
    @classmethod
    def _create_empty(cls, entity_id):
        """
        Creates a new eula instance with it's attribute set to their default values.
        
        Parameters
        ----------
        entity_id : `int`
            The entity's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.content = None
        self.id = entity_id
        self.name = ''
        return self
    

    @classmethod
    def precreate(cls, eula_id, **keyword_parameters):
        """
        Creates an eula entity instance. When the eula is loaded with the same id, it will be picked up.
        
        Parameters
        ----------
        eula_id : `int`
            The eula's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        content : `str`, Optional (Keyword only)
            The eula's content.
        
        name : `str`, Optional (Keyword only)
            The eula's name.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        eula_id = validate_id(eula_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = EULAS[eula_id]
        except KeyError:
            self = cls._create_empty(eula_id)
            EULAS[eula_id] = self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def __repr__(self):
        """Returns the eula's representation"""
        repr_parts = ['<', self.__class__.__name__]
        
        eula_id = self.id
        if eula_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two eulas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two eulas are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # content
        if self.content != other.content:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    

    
    def __hash__(self):
        """Returns the hash value of the eula."""
        hash_value = 0
        
        # content
        content = self.content
        if (content is not None):
            hash_value ^= hash(content)
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the eula.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.content = self.content
        new.id = 0
        new.name = self.name
        return new
    
    
    def copy_with(self, *, content = ..., name = ...):
        """
        Copies the eula with the defined fields.
        
        Parameters
        ----------
        content : `None`, `str`, Optional (Keyword only)
            The eula's content.
        
        name : `str`, Optional (Keyword only)
            The eula's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # content
        if content is ...:
            content = self.content
        else:
            content = validate_content(content)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        new = object.__new__(type(self))
        new.content = content
        new.id = 0
        new.name = name
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the entity is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)



from .discovery import *
from .discovery_category import *
from .embedded_activity_state import *
from .guild import *
from .guild_join_request import *
from .guild_join_request_delete_event import *
from .guild_join_request_form_response import *
from .guild_preview import *
from .guild_user_chunk_event import *
from .guild_widget import *
from .guild_widget_channel import *
from .guild_widget_user import *
from .verification_screen import *
from .verification_screen_step import *
from .welcome_screen import *
from .welcome_screen_channel import *


__all__ = (
    *discovery.__all__,
    *discovery_category.__all__,
    *embedded_activity_state.__all__,
    *guild.__all__,
    *guild_join_request.__all__,
    *guild_join_request_delete_event.__all__,
    *guild_join_request_form_response.__all__,
    *guild_preview.__all__,
    *guild_user_chunk_event.__all__,
    *guild_widget.__all__,
    *guild_widget_channel.__all__,
    *guild_widget_user.__all__,
    *verification_screen.__all__,
    *verification_screen_step.__all__,
    *welcome_screen.__all__,
    *welcome_screen_channel.__all__,
)



from .application import *
from .application_executable import *
from .application_install_parameters import *
from .application_role_connection import *
from .application_role_connection_metadata import *
from .eula import *
from .team import *
from .team_member import *
from .third_party_sku import *

from .flags import *
from .preinstanced import *


__all__ = (
    *application.__all__,
    *application_executable.__all__,
    *application_install_parameters.__all__,
    *application_role_connection.__all__,
    *application_role_connection_metadata.__all__,
    *eula.__all__,
    *team.__all__,
    *team_member.__all__,
    *third_party_sku.__all__,
    
    *flags.__all__,
    *preinstanced.__all__,
)



from .scheduled_event import *
from .scheduled_event_entity_metadata import *
from .scheduled_event_subscribe_event import *
from .scheduled_event_unsubscribe_event import *


__all__ = (
    *scheduled_event.__all__,
    *scheduled_event_entity_metadata.__all__,
    *scheduled_event_subscribe_event.__all__,
    *scheduled_event_unsubscribe_event.__all__,
)



__all__ = (
    'ALLOW_DEBUG_MESSAGES', 'API_VERSION', 'CACHE_PRESENCE', 'CACHE_USER', 'CUSTOM_API_ENDPOINT', 'CUSTOM_CDN_ENDPOINT',
    'CUSTOM_DISCORD_ENDPOINT', 'CUSTOM_STATUS_ENDPOINT', 'DOCS_ENABLED', 'LIBRARY_AGENT_APPENDIX', 'LIBRARY_NAME',
    'LIBRARY_URL', 'LIBRARY_VERSION', 'MESSAGE_CACHE_SIZE', 'RICH_DISCORD_EXCEPTION'
)

import warnings

from .getters import get_bool_env, get_int_env, get_str_env
from .loading import find_dot_env_file, load_dot_env_from_file

# Load dotenv

dot_env_file_path = find_dot_env_file()
if (dot_env_file_path is not None):
    load_dot_env_from_file(dot_env_file_path).insert_to_environmental_variables().raise_if_failed()
del dot_env_file_path

# Get variables

CACHE_PRESENCE = get_bool_env('HATA_CACHE_PRESENCE', True)
CACHE_USER = get_bool_env('HATA_CACHE_USERS', True)
MESSAGE_CACHE_SIZE = get_int_env('HATA_MESSAGE_CACHE_SIZE', 10)

if (MESSAGE_CACHE_SIZE < 0):
    MESSAGE_CACHE_SIZE = 0

DOCS_ENABLED = get_bool_env('HATA_DOCS_ENABLED', (get_bool_env is not None))
if not DOCS_ENABLED:
    get_bool_env.__doc__ = None
    get_str_env.__doc__ = None
    get_int_env.__doc__ = None

# You cannot store presences of not loaded users.
if (not CACHE_USER):
    CACHE_PRESENCE = False

ALLOW_DEBUG_MESSAGES = get_bool_env('HATA_ALLOW_DEBUG_MESSAGES', False)

CUSTOM_API_ENDPOINT = get_str_env('HATA_API_ENDPOINT')
CUSTOM_CDN_ENDPOINT = get_str_env('HATA_CDN_ENDPOINT')
CUSTOM_DISCORD_ENDPOINT = get_str_env('HATA_DISCORD_ENDPOINT')
CUSTOM_STATUS_ENDPOINT = get_str_env('HATA_STATUS_ENDPOINT')

DEFAULT_API_VERSION = 10
API_VERSION = get_int_env('HATA_API_VERSION', DEFAULT_API_VERSION)

if API_VERSION != DEFAULT_API_VERSION:
    if API_VERSION < 6:
        warnings.warn(
            f'`HATA_API_VERSION` given with a value less than `6`, got {API_VERSION!r}, defaulting to '
            f'{DEFAULT_API_VERSION!r}!'
        )
        API_VERSION = DEFAULT_API_VERSION
    
    elif API_VERSION > 11:
        warnings.warn(
            f'`API_VERSION` given with a value greater than `11`, got {API_VERSION!r}, defaulting to '
            f'{DEFAULT_API_VERSION!r}!'
        )
        API_VERSION = DEFAULT_API_VERSION
    
    elif API_VERSION < 9:
        warnings.warn(
            (
                f'`API_VERSION` given either as `6`, `7`, `8`, got {API_VERSION!r}, please use version '
                f'`{DEFAULT_API_VERSION!r}` instead',
            ),
            FutureWarning,
        )


LIBRARY_AGENT_APPENDIX = get_str_env('HATA_LIBRARY_AGENT_APPENDIX', None)
LIBRARY_NAME = get_str_env('HATA_LIBRARY_NAME', 'hata')
LIBRARY_URL = get_str_env('HATA_LIBRARY_URL', 'https://github.com/HuyaneMatsu/hata')
LIBRARY_VERSION = get_str_env('HATA_LIBRARY_VERSION', None)


RICH_DISCORD_EXCEPTION = get_bool_env('HATA_RICH_DISCORD_EXCEPTION', False)



__all__ = ()

from ...field_parsers import entity_id_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import entity_id_putter_factory, force_string_putter_factory, nullable_string_putter_factory
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_string_validator_factory
)

from .constants import DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN

# description

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('message_application_id')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



__all__ = ()

from functools import partial as partial_func

from ...field_parsers import (
    bool_parser_factory, force_string_parser_factory, int_parser_factory, nullable_functional_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, force_string_putter_factory, int_putter_factory,
    nullable_functional_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, force_string_validator_factory, int_conditional_validator_factory
)
from ...localization.helpers import localized_dictionary_builder
from ...localization.utils import build_locale_dictionary, destroy_locale_dictionary

# value

parse_value = int_parser_factory('id', 0)
put_value_into = int_putter_factory('id')
validate_value = int_conditional_validator_factory(
    'value',
    0,
    lambda value : value >= 0,
    '>= 0',
)

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', 0, 1024)

# name_localizations

parse_name_localizations = nullable_functional_parser_factory(
    'name_localizations', build_locale_dictionary
)
put_name_localizations_into = nullable_functional_optional_putter_factory(
    'name_localizations', destroy_locale_dictionary
)
validate_name_localizations = partial_func(localized_dictionary_builder, parameter_name = 'name_localizations')

# primary

parse_primary = bool_parser_factory('is_primary', False)
put_primary_into = bool_optional_putter_factory('is_primary', False)
validate_primary = bool_validator_factory('primary', False)

# Ignore additional fields:
# 
# - is_published
# - reasons_to_join
# - social_links
# - about
#
# When looking at guild discovery only `is_published` shows up, but not sure what it does.
# The rest is not even showing up, so they seem to be unused?



__all__ = ('ApplicationCommandPermission',)

import warnings

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_application_command_id, parse_application_id, parse_guild_id, parse_permission_overwrites,
    put_application_command_id_into, put_application_id_into, put_guild_id_into, put_permission_overwrites_into,
    validate_application_command_id, validate_permission_overwrites
)


class ApplicationCommandPermission(RichAttributeErrorBaseType):
    """
    Stores am ``ApplicationCommand``'s overwrites.
    
    Attributes
    ----------
    application_command_id : `int`
        The identifier of the respective ``ApplicationCommand``.
    
    application_id : `int`
        The application command's application's identifier.
    
    guild_id : `int`
        The identifier of the respective guild.
    
    permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The application command permissions overwrites relating to the respective application command in the guild.
    """
    __slots__ = ('application_command_id', 'application_id', 'guild_id', 'permission_overwrites')
    
    def __new__(cls, application_command_id, *, permission_overwrites = ...):
        """
        Creates a new ``ApplicationCommandPermission`` from the given parameters.
        
        Parameters
        ----------
        application_command_id : ``ApplicationCommand``, `int`
            The application command's identifier.
        
        permission_overwrites : `None`, `iterable` of ``ApplicationCommandPermissionOverwrite`, Optional (Keyword only)
            Overwrites for the application command.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_command
        application_command_id = validate_application_command_id(application_command_id)
        
        # application_id
        # Internal attribute
        
        # guild_id
        # Internal attribute
        
        # permission_overwrites
        if permission_overwrites is ...:
            permission_overwrites = None
        else:
            permission_overwrites = validate_permission_overwrites(permission_overwrites)
        
        self = object.__new__(cls)
        self.application_command_id = application_command_id
        self.application_id = 0
        self.guild_id = 0
        self.permission_overwrites = permission_overwrites
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermission``.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application command data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.application_command_id = parse_application_command_id(data)
        self.application_id = parse_application_id(data)
        self.guild_id = parse_guild_id(data)
        self.permission_overwrites = parse_permission_overwrites(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the application command permission to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_application_command_id_into(self.application_command_id, data, defaults)
        put_permission_overwrites_into(self.permission_overwrites, data, defaults)
        
        if include_internals:
            put_application_id_into(self.application_id, data, defaults)
            put_guild_id_into(self.guild_id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the application command permission's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # application_command_id
        repr_parts.append(' application_command_id = ')
        repr_parts.append(repr(self.application_command_id))
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            permission_overwrite_count = '0'
        else:
            permission_overwrite_count = repr(len(permission_overwrites))
        repr_parts.append(', permission overwrite count: ')
        repr_parts.append(permission_overwrite_count)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command permission's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # application_command_id
        if self.application_command_id != other.application_command_id:
            return False
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        # Unique by application_command_id
        
        # permission_overwrites
        if self.permission_overwrites != other.permission_overwrites:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command overwrite's hash value."""
        hash_value = 0
        
        # application_command_id
        hash_value ^= self.application_command_id
        
        # application_id
        # Unique by application_command_id
        
        # guild_id
        # Unique by application_command_id
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            hash_value ^= len(permission_overwrites) << 4
            
            for permission_overwrite in permission_overwrites:
                hash_value ^= hash(permission_overwrite)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application command permission.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # application_command_id
        new.application_command_id = self.application_command_id
        
        # application_id
        new.application_id = 0
        
        # guild_id
        new.guild_id = 0
        
        # permission_overwrites
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = (*permission_overwrites,)
        new.permission_overwrites = permission_overwrites
        
        return new
    
    
    def copy_with(self, application_command_id = ..., permission_overwrites = ...):
        """
        Copies the application command permission with the given fields.
        
        Parameters
        ----------
        application_command_id : ``ApplicationCommand``, `int`, Optional (Keyword only)
            The application command's identifier.
        
        permission_overwrites : `None`, `iterable` of ``ApplicationCommandPermissionOverwrite`, Optional (Keyword only)
            Overwrites for the application command.
            
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if application_command_id is ...:
            application_command_id = self.application_command_id
        else:
            application_command_id = validate_application_command_id(application_command_id)
        
        if permission_overwrites is ...:
            permission_overwrites = self.permission_overwrites
            if (permission_overwrites is not None):
                permission_overwrites = (*permission_overwrites,)
        else:
            permission_overwrites = validate_permission_overwrites(permission_overwrites)
        
        new = object.__new__(type(self))
        new.application_command_id = application_command_id
        new.application_id = 0
        new.guild_id = 0
        new.permission_overwrites = permission_overwrites
        return new
    
    
    def iter_permission_overwrites(self):
        """
        Iterates over the permission overwrites of the application command permission.
        
        This method is an iterable generator.
        
        Yields
        ------
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``
        """
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            yield from permission_overwrites
    
    
    def add_permission_overwrite(self, permission_overwrite):
        """
        Deprecated and will be removed in 2023 Jun.
        """
        warnings.warn(
            (
                f'{self.__class__.__name__} is deprecated and will be removed in 2023 Jun.'
            ),
            FutureWarning,
            stacklevel = 2,
        )



__all__ = ()

import sys
from os import getcwd as get_current_work_directory
from os.path import basename as get_file_name, dirname as get_directory_name, join as join_paths

from scarletio import any_to_any, get_short_executable

from ... import __file__ as PACKAGE_INIT_FILE, __package__ as PACKAGE_NAME


UPPER_DIRECTORY = get_directory_name(get_directory_name(PACKAGE_INIT_FILE))
PACKAGE_MAIN_FILE = join_paths(get_directory_name(PACKAGE_INIT_FILE), '__main__.py')
WORKING_DIRECTORY = get_current_work_directory()


def get_main_call(with_parameters = False):
    """
    Returns how the library main was called.
    
    Parameters
    ----------
    with_parameters : `bool` = `False`, Optional
        Whether call parameters should also be rendered.
    
    Returns
    -------
    main_call : `str`
    """
    return ''.join(render_main_call_into([], with_parameters = with_parameters))


def _render_default_main_call_into(into):
    """
    Renders the the default call way into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(get_short_executable())
    into.append(' -m ')
    into.append(PACKAGE_NAME)


def _render_parameters_into(into, parameters):
    """
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    parameters : `list` of `str`
        The parameters to render.
    
    Returns
    -------
    into : `list` of `str`
    """
    for parameter in parameters:
        into.append(' ')
        if not parameter:
            into.append('""')
            continue
        
        if not any_to_any(parameter, (' ', '"')):
            into.append(parameter)
            continue
        
        into.append('"')
        into.append(parameter.replace('"', '\\"'))
        into.append('"')
        continue

    return into


def render_main_call_into(into, with_parameters = False):
    """
    Renders how the library was called into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        List to extend.
    with_parameters : `bool` = `False`, Optional
        Whether call parameters should also be rendered.
    
    Returns
    -------
    into : `list` of `str`
    """
    system_parameters = sys.argv
    if len(system_parameters) < 1:
        into = _render_default_main_call_into(into)
    
    else:
        executed_file = system_parameters[0]
        if executed_file == PACKAGE_MAIN_FILE:
            into = _render_default_main_call_into(into)
            if with_parameters:
                into = _render_parameters_into(into, system_parameters[1:])
        
        elif (UPPER_DIRECTORY != WORKING_DIRECTORY) and (get_file_name(executed_file) == PACKAGE_NAME):
            into.append(PACKAGE_NAME)
        
        else:
            into.append(get_short_executable())
            into.append(' ')
            into.append(executed_file)
            
        if with_parameters:
            into = _render_parameters_into(into, system_parameters[1:])
    
    return into



__all__ = ('EventHandlerSnapshotType',)

from scarletio import RichAttributeErrorBaseType, copy_docs

from .helpers import _merge_list_groups, _get_list_difference
from .base_snapshot_type import BaseSnapshotType


class EventHandlerDifference(RichAttributeErrorBaseType):
    """
    Represents a difference between two event handlers.
    
    Attributes
    ----------
    added_event_handlers : `None`, `list` of `callable`
        The added event handlers.
    removed_event_handlers : `None`, `list` of `callable`
        The removed event handlers.
    """
    __slots__ = ('added_event_handlers', 'removed_event_handlers',)
    
    def __new__(cls):
        """
        Creates a new event handler difference.
        """
        self = object.__new__(cls)
        self.added_event_handlers = None
        self.removed_event_handlers = None
        
        return self
    
    
    def __repr__(self):
        """Returns the event handler difference."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' added_event_handlers=')
            repr_parts.append(repr(added_event_handlers))
        
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(',4 removed_event_handlers=')
            repr_parts.append(repr(removed_event_handlers))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __bool__(self):
        """Returns whether the difference contains any changes."""
        if (self.added_event_handlers is not None):
            return True
        
        if (self.removed_event_handlers is not None):
            return True
        
        return False
    
    
    def add(self, event_handler):
        """
        Adds an event handler to self.
        
        Parameters
        ----------
        event_handler : `callable`
            The event handler to add.
        """
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is None):
            added_event_handlers = []
            self.added_event_handlers = added_event_handlers
        
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is not None):
            try:
                removed_event_handlers.remove(event_handler)
            except ValueError:
                pass
            else:
                if not removed_event_handlers:
                    self.removed_event_handlers = None
        
        added_event_handlers.append(event_handler)
    
    
    def remove(self, event_handler):
        """
        Removes the event handler from self.
        
        Parameters
        ----------
        event_handler : `callable`
            The event handler to remove.
        """
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is not None):
            try:
                added_event_handlers.remove(event_handler)
            except ValueError:
                pass
            else:
                if not added_event_handlers:
                    self.added_event_handlers = None
        
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is None):
            removed_event_handlers = []
            self.removed_event_handlers = removed_event_handlers
        
        removed_event_handlers.append(event_handler)
    
    
    def iter_added_event_handlers(self):
        """
        Iterates over the added event handlers of the event handler difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        added_event_handler : `callable`
        """
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is not None):
            yield from added_event_handlers
    
    
    def iter_removed_event_handlers(self):
        """
        Iterates over the removed event handlers of the event handler difference.
        
        This method is an iterable generator.
        
        Yields
        ------
        removed_event_handler : `callable`
        """
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is not None):
            yield from removed_event_handlers
    
    
    def copy(self):
        """
        Copies the event handler difference.
        
        Returns
        -------
        new : ``EventHandlerDifference``
        """
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is not None):
            added_event_handlers = added_event_handlers.copy()
        
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is not None):
            removed_event_handlers = removed_event_handlers.copy()
        
        new = object.__new__(type(self))
        new.added_event_handlers = added_event_handlers
        new.removed_event_handlers = removed_event_handlers
        return new
    
    
    def revert_copy(self):
        """
        Revert copies the event handler difference.
        
        Returns
        -------
        new : ``EventHandlerDifference``
        """
        added_event_handlers = self.added_event_handlers
        if (added_event_handlers is not None):
            added_event_handlers = added_event_handlers.copy()
        
        removed_event_handlers = self.removed_event_handlers
        if (removed_event_handlers is not None):
            removed_event_handlers = removed_event_handlers.copy()
        
        new = object.__new__(type(self))
        new.added_event_handlers = removed_event_handlers
        new.removed_event_handlers = added_event_handlers
        return new
    
    
    def _subtract(self, other):
        """
        Subtracts other from self.
        
        Parameters
        ----------
        other : ``EventHandlerDifference``
            The other difference to subtract from self.
        
        Returns
        -------
        new : ``EventHandlerDifference``
        """
        added_event_handlers, removed_event_handlers = _get_list_difference(
            _merge_list_groups(self.added_event_handlers, other.removed_event_handlers),
            _merge_list_groups(other.added_event_handlers, self.removed_event_handlers),
        )
        
        new = object.__new__(type(self))
        new.added_event_handlers = added_event_handlers
        new.removed_event_handlers = removed_event_handlers
        return new
    
    
    def __sub__(self, other):
        """Subtracts other from self."""
        if other is None:
            return self.copy()
        
        if type(self) is type(other):
            return self._subtract(other)
        
        return NotImplemented
    
    
    def __rsub__(self, other):
        """subtracts other from self."""
        if other is None:
            return self.revert_copy()
        
        if type(self) is type(other):
            return other._subtract(self)
        
        return NotImplemented


class EventHandlerSnapshotType(BaseSnapshotType):
    """
    Event handler snapshot.
    
    Attributes
    ----------
    _client_reference : `WeakReferer` to `Client`
        Weakreference to the owner client instance.
    _is_difference : `bool`
        Whether the snapshot is a difference.
    differences : `dict` of (`str`, `list` of `callable`) items
        The collected event handlers.
    """
    __slots__ = ('differences',)
    
    @copy_docs(BaseSnapshotType.__new__)
    def __new__(cls, client):
        differences = {}
        
        event_handler_manager = client.events
        for event_handler_name, event_handler in event_handler_manager.iter_event_names_and_handlers():
            try:
                difference = differences[event_handler_name]
            except KeyError:
                difference = EventHandlerDifference()
                differences[event_handler_name] = difference
            
            difference.add(event_handler)
        
        self = BaseSnapshotType.__new__(cls, client)
        self.differences = differences
        return self
    
    
    @copy_docs(BaseSnapshotType._extract)
    def _extract(self, other):
        new_differences = {}
        
        self_differences = self.differences
        other_difference = other.differences
        
        for event_handler_name in {*self_differences.keys(), *other_difference.keys()}:
            new_difference = \
                self_differences.get(event_handler_name, None) - other_difference.get(event_handler_name, None)
            
            if new_difference:
                new_differences[event_handler_name] = new_difference
        
        new = BaseSnapshotType._extract(self, other)
        new.differences = new_differences
        return new
    
    
    @copy_docs(BaseSnapshotType.revert)
    def revert(self):
        BaseSnapshotType.revert(self)
        
        client = self.client
        if client is None:
            return
        
        
        event_handler_manager = client.events
        for event_handler_name, difference in self.differences.items():
            for event_handler in difference.iter_added_event_handlers():
                event_handler_manager.remove(event_handler, name = event_handler_name, count = 1)
            
            for event_handler in difference.iter_removed_event_handlers():
                event_handler_manager(event_handler, name = event_handler_name)
    
    
    @copy_docs(BaseSnapshotType.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        client = self.client
        if (client is not None):
            repr_parts.append(' of ')
            repr_parts.append(repr(client))
        
        
        differences = self.differences
        if differences:
            repr_parts.append(', differences = ')
            repr_parts.append(repr(self.differences))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(BaseSnapshotType.__bool__)
    def __bool__(self):
        if self.differences:
            return True
        
        return False



__all__ = ('get_bool_env', 'get_int_env', 'get_str_env',)


import warnings
from os import getenv as get_environmental_variable


ERROR_MESSAGE_APPENDIX = 'Please update your `.env` file.'


def get_bool_env(name, default = False, *, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or not present as a boolean's representation returns `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `bool` = `False`, Optional
        The default value of the respective variable.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    env_variable : `bool`
    """
    env_variable = get_environmental_variable(name)
    if env_variable is None:
        return default
    
    env_variable = env_variable.casefold()
    if env_variable in ('true', '0'):
        return True
    
    if env_variable in ('false', '0'):
        return False
    
    if warn_if_empty:
        warnings.warn(
            f'{name!r} is specified as non bool: {env_variable!r}, defaulting to {default!r}!',
        )
    
    return default


def get_str_env(name, default = None, *, raise_if_missing_or_empty = False, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or present as an empty string returns `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `None` | `str` = `None`, Optional
        The default value of the respective variable. Defaults to `None`.
    raise_if_missing_or_empty : `bool` = `False`, Optional (Keyword only)
        Whether exception should be thrown if the environmental variable is missing or empty.
        Has priority over `default` and `warn_if_empty`.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    variable : `None` | `str`
    
    Raises
    ------
    RuntimeError
    """
    env_variable = get_environmental_variable(name)
    if env_variable is None:
        if raise_if_missing_or_empty:
            raise RuntimeError(
                f'Environmental variable {name!r} is missing. {ERROR_MESSAGE_APPENDIX}'
            )
        
        return default
    
    if env_variable:
        return env_variable
    
    if raise_if_missing_or_empty:
        raise RuntimeError(
            f'Environmental variable {name!r} is specified as empty string. {ERROR_MESSAGE_APPENDIX}'
        )
    
    if warn_if_empty:
        warnings.warn(
            f'{name!r} is specified as empty string, defaulting to {default!r}!',
        )
    
    return default


def get_int_env(name, default = 0, *, warn_if_empty = True):
    """
    Gets the given environmental variable.
    
    If the environmental variable is not present or present as non `int`, will return `default` instead.
    
    Parameters
    ----------
    name : `str`
        The name of an environmental variable.
    default : `int` = `0`, Optional
        The default value of the respective variable.
    warn_if_empty : `bool` = `True`, Optional (Keyword only)
        Whether warning should be dropped if empty environmental variable is received.
    
    Returns
    -------
    variable : `int`
    """
    env_variable = get_environmental_variable(name)
    if env_variable is None:
        return default
    
    if not env_variable:
        return default
    
    try:
        variable = int(env_variable)
    except ValueError:
        pass
    else:
        return variable
    
    if warn_if_empty:
        warnings.warn(
            f'{name!r} is specified as non `int`: {env_variable!r}, defaulting to {default!r}!',
        )
    
    return default




__all__ = ('ActivityUpdate',)

from itertools import count

from scarletio import RichAttributeErrorBaseType

from ...activity import Activity

from .fields import validate_activity, validate_old_attributes


class ActivityUpdate(RichAttributeErrorBaseType):
    """
    Represents an updated activity with storing the activity and it's old updated attributes in a `dict`.
    
    Attributes
    ----------
    activity : ``Activity``
        The updated activity.
    old_attributes : `dict` of (`str`, `object`) items
        The changed attributes of the activity in `attribute-name` - `old-value` relation. Can contain any of the
        following items:
        
        +-------------------+-----------------------------------+
        | Keys              | Values                            |
        +===================+===================================+
        | assets            | `None`, ``ActivityAssets``        |
        +-------------------+-----------------------------------+
        | created_at        | `datetime`                        |
        +-------------------+-----------------------------------+
        | details           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | emoji             | `None`, ``Emoji``                 |
        +-------------------+-----------------------------------+
        | flags             | ``ActivityFlag``                  |
        +-------------------+-----------------------------------+
        | name              | `str`                             |
        +-------------------+-----------------------------------+
        | metadata          | ``ActivityMetadataBase``          |
        +-------------------+-----------------------------------+
        | party             | `None`, ``ActivityParty``         |
        +-------------------+-----------------------------------+
        | secrets           | `None`, ``ActivitySecrets``       |
        +-------------------+-----------------------------------+
        | session_id        | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | state             | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | sync_id           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | timestamps        | `None`, `ActivityTimestamps``     |
        +-------------------+-----------------------------------+
        | url               | `None`, `str`                     |
        +-------------------+-----------------------------------+
    """
    __slots__ = ('activity', 'old_attributes',)
    
    def __new__(cls, activity = ..., old_attributes = ...):
        """
        Creates a new activity change instance with the given fields.
        
        Parameters
        ----------
        activity : ``Activity``, Optional (Keyword only)
            The updated activity.
        old_attributes : `None`, `dict` of (`str`, `object`) items, Optional (Keyword only)
            The changed attributes of the activity.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # activity
        if activity is ...:
            activity = Activity()
        else:
            activity = validate_activity(activity)
        
        # old_attributes
        if old_attributes is ...:
            old_attributes = {}
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # Construct
        self = object.__new__(cls)
        self.activity = activity
        self.old_attributes = old_attributes
        return self
    
    
    @classmethod
    def from_fields(cls, activity, old_attributes):
        """
        Creates a new activity change instance with the given fields.
        
        Parameters
        ----------
        activity : ``Activity``
            The updated activity.
        old_attributes : `dict` of (`str`, `object`) items
            The changed attributes of the activity.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.activity = activity
        self.old_attributes = old_attributes
        return self
    
    
    def __repr__(self):
        """Returns the representation of the activity update."""
        return f'<{self.__class__.__name__} changes: {len(self.old_attributes)!r}, activity = {self.activity!r}>'
    
    
    def __hash__(self):
        """Returns the activity update's hash value."""
        hash_value = 0
        
        # activity
        hash_value ^= hash(self.activity)
        
        # old_attributes
        old_attributes = self.old_attributes
        hash_value ^= len(old_attributes)
        
        for mask, key in zip(count(5, 7), sorted(old_attributes.keys())):
            hash_value ^= (mask | hash(key)) & hash(old_attributes[key])
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.activity != other.activity:
            return False
        
        if self.old_attributes != other.old_attributes:
            return False
        
        return True
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 2
    
    
    def __iter__(self):
        """
        Unpacks the activity update.
        
        This method is a generator.
        """
        yield self.activity
        yield self.old_attributes
    
    
    def copy(self):
        """
        Copies the activity update.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.activity = self.activity.copy()
        new.old_attributes = self.old_attributes.copy()
        return new
    
    
    def copy_with(self, *, activity = ..., old_attributes = ...):
        """
        Copies the activity change with the given fields.
        
        Parameters
        ----------
        activity : ``Activity``, Optional (Keyword only)
            The updated activity.
        old_attributes : `None`, `dict` of (`str`, `object`) items, Optional (Keyword only)
            The changed attributes of the activity.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # activity
        if activity is ...:
            activity = self.activity.copy()
        else:
            activity = validate_activity(activity)
        
        # old_attributes
        if old_attributes is ...:
            old_attributes = self.old_attributes.copy()
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # Construct
        new = object.__new__(type(self))
        new.activity = activity
        new.old_attributes = old_attributes
        return new



__all__ = ('PLUGINS',)

import re
from scarletio import WeakValueDictionary


PLUGINS = WeakValueDictionary()

PLUGIN_STATE_UNDEFINED = 0
PLUGIN_STATE_LOADED = 1
PLUGIN_STATE_UNLOADED = 2
PLUGIN_STATE_UNSATISFIED = 3

PLUGIN_STATE_VALUE_TO_NAME = {
    PLUGIN_STATE_UNDEFINED: 'undefined',
    PLUGIN_STATE_LOADED: 'loaded',
    PLUGIN_STATE_UNLOADED: 'unloaded',
    PLUGIN_STATE_UNSATISFIED: 'unsatisfied',
}

LOADING_PLUGINS = set()

# Plugin roots define core part for plugins.
PLUGIN_ROOTS = set()


IGNORED_DIRECTORY_NAMES = frozenset((
    '__pycache__',
    'tests',
))


ABSOLUTE_PATH_PLUGIN_NAME_PREFIX = '<unknown>.'

IN_DIRECTORY_PLUGIN_RP = re.compile('\\.[^.]+')


PLUGIN_ACTION_FLAG_NONE = 0
PLUGIN_ACTION_FLAG_LOAD = 1 << 0
PLUGIN_ACTION_FLAG_UNLOAD = 1 << 1
PLUGIN_ACTION_FLAG_NAME_LOOKUP = 1 << 2
PLUGIN_ACTION_FLAG_SYNTAX_CHECK = 1 << 3
PLUGIN_ACTION_FLAG_UNLINK = 1 << 4



__all__ = ('ActivityType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..activity_metadata import ActivityMetadataBase, ActivityMetadataCustom, ActivityMetadataRich


class ActivityType(PreinstancedBase):
    """
    Represents an ``AutoModerationAction``'s type.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the activity type.
    name : `str`
        The default name of the activity type.
    metadata_type : `type<ActivityMetadataBase>`
        The activity type's respective metadata type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``ActivityType``) items
        Stores the predefined activity types. This container is accessed when translating a Discord side
        identifier of a activity type. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `type` = `str`
        The activity types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the activity types.
    
    Every predefined activity type is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | Class attribute name  | Name                  | Value     | Metadata type                         |
    +=======================+=======================+===========+=======================================+
    | unknown               | unknown               | -1        | ``ActivityMetadataBase``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | game                  | game                  | 0         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | stream                | block stream          | 1         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | spotify               | spotify               | 2         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | watching              | watching              | 3         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | custom                | custom                | 4         | ``ActivityMetadataCustom``            |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    | competing             | competing             | 5         | ``ActivityMetadataRich``              |
    +-----------------------+-----------------------+-----------+---------------------------------------+
    """
    __slots__ = ('metadata_type',)
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new activity type with the given value.
        
        Parameters
        ----------
        value : `int`
            The activity type's identifier value.
        
        Returns
        -------
        self : ``ActivityType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.metadata_type = ActivityMetadataRich
        
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates an ``ActivityType`` and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the activity type.
        name : `str`
            The default name of the activity type.
        metadata_type : `None`, `type<ActivityMetadataBase>`
            The activity type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    # predefined
    unknown = P(-1, 'unknown', ActivityMetadataBase)
    game = P(0, 'game', ActivityMetadataRich)
    stream = P(1, 'stream message', ActivityMetadataRich)
    spotify = P(2, 'spotify', ActivityMetadataRich)
    watching = P(3, 'watching', ActivityMetadataRich)
    custom = P(4, 'custom', ActivityMetadataCustom)
    competing = P(5, 'competing', ActivityMetadataRich)



from .base import *
from .custom import *
from .flags import *
from .fields import *
from .rich import *


__all__ = (
    *base.__all__,
    *custom.__all__,
    *flags.__all__,
    *fields.__all__,
    *rich.__all__,
)



__all__ = ()

from ..channel import Channel
from ..field_parsers import (
    bool_parser_factory, entity_id_parser_factory, negated_bool_parser_factory, nullable_string_parser_factory,
    preinstanced_parser_factory
)
from ..field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, entity_id_putter_factory,
    negated_bool_optional_putter_factory, nullable_string_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ..field_validators import (
    bool_validator_factory, entity_id_validator_factory, nullable_string_validator_factory,
    preinstanced_validator_factory
)
from ..scheduled_event import PrivacyLevel
from ..scheduled_event import ScheduledEvent

from .constants import TOPIC_LENGTH_MAX, TOPIC_LENGTH_MIN


# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', Channel)

# discoverable

parse_discoverable = negated_bool_parser_factory('discoverable_disabled', True)
put_discoverable_into = negated_bool_optional_putter_factory('discoverable_disabled', True)
validate_discoverable = bool_validator_factory('discoverable', True)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('emoji_id')

# invite_code

parse_invite_code = nullable_string_parser_factory('invite_code')
put_invite_code_into = url_optional_putter_factory('invite_code')
validate_invite_code = nullable_string_validator_factory('invite_code', 0, 1024)

# privacy_level

parse_privacy_level = preinstanced_parser_factory('privacy_level', PrivacyLevel, PrivacyLevel.guild_only)
put_privacy_level_into = preinstanced_putter_factory('privacy_level')
validate_privacy_level = preinstanced_validator_factory('privacy_level', PrivacyLevel)

# scheduled_event_id

parse_scheduled_event_id = entity_id_parser_factory('guild_scheduled_event_id')
put_scheduled_event_id_into = entity_id_optional_putter_factory('guild_scheduled_event_id')
validate_scheduled_event_id = entity_id_validator_factory('scheduled_event_id', ScheduledEvent)

# send_start_notification

parse_send_start_notification = bool_parser_factory('send_start_notification', False)
put_send_start_notification_into = bool_optional_putter_factory('send_start_notification', False)
validate_send_start_notification = bool_validator_factory('send_start_notification', False)

# topic

parse_topic = nullable_string_parser_factory('topic')
put_topic_into = nullable_string_putter_factory('topic')
validate_topic = nullable_string_validator_factory('topic', TOPIC_LENGTH_MIN, TOPIC_LENGTH_MAX)



__all__ = ()

from .parameter import TYPE_IDENTIFIER_BOOL, TYPE_IDENTIFIER_FLOAT, TYPE_IDENTIFIER_INT, TYPE_IDENTIFIER_STR


PARAMETER_NOTE_SIGN_REQUIRED = '*'
PARAMETER_NOTE_SIGN_MULTIPLE_POSITIONAL = '**'
PARAMETER_NOTE_SIGN_MULTIPLE_KEYWORD = '***'

PARAMETER_NOTE_ADJUSTMENT = max(
    len(PARAMETER_NOTE_SIGN_REQUIRED),
    len(PARAMETER_NOTE_SIGN_MULTIPLE_POSITIONAL),
    len(PARAMETER_NOTE_SIGN_MULTIPLE_KEYWORD),
) + 1

PARAMETER_DEFAULT_PREFIX = 'Default: '

PARAMETER_COLUMN_SEPARATOR = '   '
PARAMETER_MODIFIER_SEPARATOR = ' / '

PARAMETER_TYPE_IDENTIFIER_TO_REPRESENTATION = {
    TYPE_IDENTIFIER_STR: 'TEXT',
    TYPE_IDENTIFIER_INT: 'INTEGER',
    TYPE_IDENTIFIER_FLOAT: 'FLOAT',
    TYPE_IDENTIFIER_BOOL: 'BOOLEAN'
}

BOX_LEFT_TOP = '╭─'
BOX_RIGHT_TOP = '─╮'
BOX_LEFT_BOT = '╰─'
BOX_RIGHT_BOT = '─╯'

BOX_LEFT = '│ '
BOX_RIGHT = ' │'

BOX_TOP = '─'
BOX_BOTTOM = '─'

BOX_TITLE_PARAMETER = ' PARAMETERS '
BOX_TITLE_MODIFIER = ' OPTIONS '
BOX_TITLE_SUB_COMMANDS = ' SUB-COMMANDS '
BOX_TITLE_ERROR = ' ERROR '

NOTE_SIGN_DESCRIPTION = {
    PARAMETER_NOTE_SIGN_REQUIRED: 'Required.',
    PARAMETER_NOTE_SIGN_MULTIPLE_POSITIONAL: 'Allows multiple parameters.',
    PARAMETER_NOTE_SIGN_MULTIPLE_KEYWORD: 'Captures any keyword parameters.'
}



__all__ = ()

from ...field_parsers import (
    default_entity_parser_factory, entity_id_parser_factory, force_string_parser_factory,
    nullable_entity_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    default_entity_putter_factory, entity_id_optional_putter_factory, force_string_putter_factory,
    nullable_entity_optional_putter_factory, preinstanced_putter_factory
)
from ...field_validators import (
    default_entity_validator, entity_id_validator_factory, force_string_validator_factory,
    nullable_entity_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

from ..webhook_source_channel import WebhookSourceChannel
from ..webhook_source_guild import WebhookSourceGuild

from .preinstanced import WebhookType

# application_id

parse_application_id = entity_id_parser_factory('application_id')
put_application_id_into = entity_id_optional_putter_factory('application_id')
validate_application_id = entity_id_validator_factory('application_id', NotImplemented, include = 'Application')

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_optional_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', NotImplemented, include = 'Channel')

# source_channel

parse_source_channel = nullable_entity_parser_factory('source_channel', WebhookSourceChannel)
put_source_channel_into = nullable_entity_optional_putter_factory('source_channel', WebhookSourceChannel)
validate_source_channel = nullable_entity_validator_factory('source_channel', WebhookSourceChannel)

# source_guild

parse_source_guild = nullable_entity_parser_factory('source_guild', WebhookSourceGuild)
put_source_guild_into = nullable_entity_optional_putter_factory('source_guild', WebhookSourceGuild)
validate_source_guild = nullable_entity_validator_factory('source_guild', WebhookSourceGuild)

# token

parse_token = force_string_parser_factory('token')
put_token_into = force_string_putter_factory('token')
validate_token = force_string_validator_factory('token', 0, 4096)

# type

parse_type = preinstanced_parser_factory('type', WebhookType, WebhookType.none)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('webhook_type', WebhookType)

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user_into = default_entity_putter_factory('user', ClientUserBase, ZEROUSER)
validate_user = default_entity_validator('user', ClientUserBase, default = ZEROUSER)



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, default_entity_parser_factory, entity_id_parser_factory, force_string_parser_factory,
    int_parser_factory, nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, default_entity_putter_factory, entity_id_optional_putter_factory,
    entity_id_putter_factory, force_string_putter_factory, int_putter_factory, nullable_string_putter_factory, 
    preinstanced_putter_factory
)
from ...field_validators import (
    bool_validator_factory, default_entity_validator, entity_id_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_string_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

from .constants import (
    DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN, SORT_VALUE_DEFAULT,
    SORT_VALUE_MAX, SORT_VALUE_MIN
)
from .preinstanced import StickerFormat, StickerType

# available

parse_available = bool_parser_factory('available', True)
put_available_into = bool_optional_putter_factory('available', True)
validate_available = bool_validator_factory('available', True)

# description 

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX)

# format

parse_format = preinstanced_parser_factory('format_type', StickerFormat, StickerFormat.none)
put_format_into = preinstanced_putter_factory('format_type')
validate_format = preinstanced_validator_factory('format', StickerFormat)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# pack_id

parse_pack_id = entity_id_parser_factory('pack_id')
put_pack_id_into = entity_id_optional_putter_factory('pack_id')
validate_pack_id = entity_id_validator_factory('pack_id', NotImplemented, include = 'StickerPack')

# sort_value

parse_sort_value = int_parser_factory('sort_value', SORT_VALUE_DEFAULT)
put_sort_value_into = int_putter_factory('sort_value')
validate_sort_value = int_conditional_validator_factory(
    'sort_value',
    SORT_VALUE_MIN,
    (lambda sort_value: sort_value >= SORT_VALUE_MIN and sort_value <= SORT_VALUE_MAX),
    f'>= {SORT_VALUE_MIN} and <= {SORT_VALUE_MAX},'
)

# tags

def parse_tags(data):
    """
    Parses out sticker tags from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Sticker data
    
    Returns
    -------
    tags : `None`, `frozenset` of `str`
    """
    raw_tags = data.get('tags', None)
    if (raw_tags is None) or (not raw_tags):
        tags = None
    else:
        tags = frozenset(raw_tags.split(', '))

    return tags


def put_tags_into(tags, data, defaults):
    """
    Puts the sticker tags into the given json serializable dictionary.
    
    Parameters
    ----------
    tags : `None`, `frozenset` of `str`
        Sticker tags.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if tags is None:
        raw_tags = ''
    else:
        raw_tags = ', '.join(tags)
    data['tags'] = raw_tags
    return data


def validate_tags(tags):
    """
    Validates the given sticker tags.
    
    Parameters
    ----------
    tags : `None`, `str`, `iterable` of `str`
        Sticker tags to validate.
    
    Returns
    -------
    tags : `None`, `frozenset` of `str`
    
    Raises
    ------
    TypeError
        - If `tags`'s value or structure is incorrect.
    """
    if tags is None:
        return tags
    
    if isinstance(tags, str):
        if not tags:
            return None
        
        return frozenset(tags.split(', '))
    
    if getattr(type(tags), '__iter__', None) is None:
        raise TypeError(
            f'`tags` can be `None`, `str`, `iterable` of `str`, got {type(tags).__name__}; {tags!r}.'
        )
    
    validated_tags = None
    
    for tag in tags:
        if not isinstance(tag, str):
            raise TypeError(
                f'`tags` can contain `str` instances, got {tag.__class__.__name__}; {tag!r}; tags = {tags!r}.'
            )
        
        if validated_tags is None:
            validated_tags = set()
        
        validated_tags.add(tag)
    
    if (validated_tags is not None):
        return frozenset(tags)

# type

parse_type = preinstanced_parser_factory('type', StickerType, StickerType.none)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('sticker_type', StickerType)

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user_into = default_entity_putter_factory('user', ClientUserBase, ZEROUSER)
validate_user = default_entity_validator('user', ClientUserBase, default = ZEROUSER)



__all__ = ('CommandResult',)

from math import floor, log10

from scarletio import RichAttributeErrorBaseType, render_exception_into

from ..helpers import render_main_call_into

from ..constants import REGISTERED_COMMANDS

from .helpers import command_sort_key
from .parameter import TYPE_IDENTIFIER_TO_NAME
from .rendering_helpers import render_error_box_into_multi_line, render_error_box_into_single_line


class CommandResult(RichAttributeErrorBaseType):
    """
    Represents an command's result.
    
    Attributes
    ----------
    error_code : `int`
        Command result code.
    detail_parameters : `tuple` of `str`
        Additional parameters to pass to the result processor.
    """
    __slots__ = ('detail_parameters', 'error_code',)
    
    def __new__(cls, error_code, *detail_parameters):
        """
        Creates a new ``CommandResult`` with the given parameters.
        
        Parameters
        ----------
        error_code : `int`
            Command result code.
        *detail_parameters : Positional parameters
            Additional parameters to pass to the result processor.
        """
        self = object.__new__(cls)
        self.error_code = error_code
        self.detail_parameters = detail_parameters
        return self
    
    
    def get_message(self):
        """
        Returns the command's result message.
        
        Returns
        -------
        message : `str`
        """
        return COMMAND_RESULT_CODE_TO_MESSAGE_PROCESSOR[self.error_code](*self.detail_parameters)


COMMAND_RESULT_CODE_CONVERSION_FAILED = 1
COMMAND_RESULT_CODE_PARAMETER_REQUIRED = 2
COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED = 3
COMMAND_RESULT_CODE_PARAMETER_EXTRA = 4
COMMAND_RESULT_CODE_PARAMETER_UNSATISFIED = 5
COMMAND_RESULT_CODE_COMMAND_UNINITIALIZED = 6
COMMAND_RESULT_CODE_CATEGORY_EMPTY = 7
COMMAND_RESULT_CODE_CATEGORY_REQUIRES_PARAMETER = 8
COMMAND_RESULT_CODE_CATEGORY_UNKNOWN_SUB_COMMAND = 9
COMMAND_RESULT_CODE_COMMAND_REQUIRED = 10
COMMAND_RESULT_CODE_CALL = 11
COMMAND_RESULT_CODE_COMMAND_NOT_FOUND = 12


def command_result_processor_conversion_failed(command_line_parameter, received_value):
    """
    Command result message processor if conversion fails.
    
    Parameters
    ----------
    command_line_parameter : ``CommandParameter``
        Respective command parameter.
    received_value : `str`
        Received value, which could not be converted to the respective type.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Parameter {command_line_parameter.display_name!r} received value of incorrect type.',
    )
    
    message_parts.append('\nReceived value: ')
    message_parts.append(repr(received_value))
    message_parts.append('\n\nExpected type: ')
    
    type_name = TYPE_IDENTIFIER_TO_NAME[command_line_parameter.expected_type_identifier]
    message_parts.append(type_name)
    message_parts.append('\n')
    
    return ''.join(message_parts)


def command_result_processor_parameter_required(command_line_parameter):
    """
    Command result message processor if a parameter stays unsatisfied.
    
    Parameters
    ----------
    command_line_parameter : ``CommandParameter``
        Respective command parameter.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Parameter {command_line_parameter.display_name!r} is required.',
    )
    
    return ''.join(message_parts)


def command_result_processor_parameter_unexpected(parameter_value):
    """
    Command result message processor if a received parameter is unexpected.
    
    Parameters
    ----------
    parameter_value : `str`
        Received parameter.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Unexpected parameter {parameter_value!r}.',
    )
    
    return ''.join(message_parts)


def command_result_processor_parameter_extra(parameter_values):
    """
    Command result message processor if received extra parameter(s).
    
    Parameters
    ----------
    parameter_values : `list` of `str`
        Extra parameters.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Extra parameter(s).',
    )
    
    limit = len(parameter_values)
    if limit:
        index = 0
        index_adjust = floor(log10(limit)) + 1
        
        while True:
            parameter_value = parameter_values[index]
            index += 1
            
            message_parts.append(str(index).rjust(index_adjust))
            message_parts.append(repr(parameter_value))
            
            if index == limit:
                break
            
            message_parts.append('\n')
            continue
    
    return ''.join(message_parts)


def command_result_processor_parameter_unsatisfied(command_line_parameter):
    """
    Command result message processor if a parameter name is defined, but it's value is not.
    
    Parameters
    ----------
    command_line_parameter : ``CommandParameter``
        Respective command parameter.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Parameter {command_line_parameter.display_name!r} is unsatisfied.',
    )
    
    return ''.join(message_parts)


def command_result_processor_command_uninitialized(command_line_command):
    """
    Command result message processor if a command line command is not initialized (should not happen).
    
    Parameters
    ----------
    command_line_command : ``Command``
        Respective command.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(
        message_parts,
        f'Command {command_line_command.name!r} is not initialized correctly.',
    )
    
    return ''.join(message_parts)
    

def command_result_processor_category_empty(command_category):
    """
    Command result message processor if a command category is empty.
    
    Parameters
    ----------
    command_category : ``CommandCategory``
        Respective command category.
    
    Returns
    -------
    message : `str`
    """
    command_full_name = ' '.join(command_category._trace_back_name())
    
    message_parts = []
    
    message_parts = render_error_box_into_multi_line(
        message_parts,
        [
            f'Command {command_full_name!r} cannot be invoked.',
            f'There are no available sub-commands either.',
        ],
    )
    
    return ''.join(message_parts)


def command_result_processor_category_requires_parameter(command_category):
    """
    Command result message processor if a command category has no command function to call, when no extra parameter
    is defined.
    
    Parameters
    ----------
    command_category : ``CommandCategory``
        Respective command category.
    
    Returns
    -------
    message : `str`
    """
    command_full_name = ' '.join(command_category._trace_back_name())
    sub_command_full_names = {
        command_function.get_full_name() for command_function in command_category.iter_command_functions()
    }
    
    sub_command_full_names.discard(command_full_name)
    
    if not sub_command_full_names:
        return command_result_processor_category_empty(command_category)
    
    message_parts = []
    
    message_parts = render_error_box_into_multi_line(
        message_parts,
        [
            f'Command {command_full_name!r} cannot be invoked.',
            f'Please define a sub command by name to call.',
        ],
    )
    
    message_parts.append(
        '\nAvailable sub-commands:\n'
    )
    
    for sub_command_full_name in sub_command_full_names:
        message_parts.append('- ')
        message_parts.append(sub_command_full_name)
        message_parts.append('\n')
    
    return ''.join(message_parts)


def command_result_processor_category_unknown_sub_command(command_category, command_name):
    """
    Command result message processor if a command category received unknown sub-command.
    
    Parameters
    ----------
    command_category : ``CommandCategory``
        Respective command category.
    command_name : `str`
        The received command name.
    
    Returns
    -------
    message : `str`
    """
    command_full_name = ' '.join(command_category._trace_back_name())

    message_parts = []
    message_parts = render_error_box_into_single_line(
        message_parts, f'Command category: {command_full_name!r} has no sub command: {command_name!r}.'
    )
    message_parts.append(
        '\n'
        'Available sub-commands:\n'
    )
    
    command_categories = command_category._command_categories
    if (command_categories is not None):
        for command_category_name in sorted(command_categories.keys()):
            message_parts.append('- ')
            message_parts.append(command_category_name)
            message_parts.append('\n')
    
    return ''.join(message_parts)


def command_result_processor_command_required():
    """
    Command result message processor if a command name was not given to run.
    
    Returns
    -------
    message : `str`
    """
    return ''.join(render_error_box_into_single_line([], 'Command name required.'))



def _ignore_command_call_frame(file_name, name, line_number, line):
    """
    Ignores the frame where the command was called.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line_number : `int`
        The line's index where the exception occurred.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    if file_name == __file__:
        if name == 'command_result_processor_call':
            if line == 'result = function(*positional_parameters, **keyword_parameters)':
                should_show_frame = False
    
    return should_show_frame


def command_result_processor_call(command_function, positional_parameters, keyword_parameters):
    """
    Parameters
    ----------
    command_function : ``CommandFunction``
        Command function to call it's function of.
    positional_parameters : `list` of `Any`
        Positional parameters to call the respective function with.
    keyword_parameters : `dict` of (`str`, `Any`) items
        Keyword parameters to call the respective function with.

    Returns
    -------
    message : `str`
    
    Raises
    ------
    SystemExit
    """
    try:
        result = command_function._function(*positional_parameters, **keyword_parameters)
    except (KeyboardInterrupt, SystemExit):
        raise
    
    except BaseException as err:
        result = ''.join(
            render_exception_into(
                err,
                [
                    'Exception occurred while invoking "',
                    command_function.get_full_name(),
                    '".\n'
                ],
                filter = _ignore_command_call_frame,
            ),
        )
    
    else:
        if (result is not None) and (not isinstance(result, str)):
            result = str(result)
    
    return result


def command_result_processor_command_not_found(command_name):
    """
    Command result message processor if a command name was not given to run.
    
    Parameters
    ----------
    command_name : `str`
        Command name.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts = render_error_box_into_single_line(message_parts, f'No command for name: {command_name!r}.')
    
    message_parts.append(
        '\n'
        'The available commands are the following:\n'
    )
    
    for command in sorted(REGISTERED_COMMANDS, key = command_sort_key):
        message_parts.append('- ')
        message_parts.append(command.name)
        message_parts.append('\n')
    
    message_parts.append('\nTry using "$ ')
    render_main_call_into(message_parts)
    message_parts.append(' help COMMAND-NAME" for more information.\n')
    
    return ''.join(message_parts)


COMMAND_RESULT_CODE_TO_MESSAGE_PROCESSOR = {
    COMMAND_RESULT_CODE_CONVERSION_FAILED: command_result_processor_conversion_failed,
    COMMAND_RESULT_CODE_PARAMETER_REQUIRED: command_result_processor_parameter_required,
    COMMAND_RESULT_CODE_PARAMETER_UNEXPECTED: command_result_processor_parameter_unexpected,
    COMMAND_RESULT_CODE_PARAMETER_EXTRA: command_result_processor_parameter_extra,
    COMMAND_RESULT_CODE_PARAMETER_UNSATISFIED: command_result_processor_parameter_unsatisfied,
    COMMAND_RESULT_CODE_COMMAND_UNINITIALIZED: command_result_processor_command_uninitialized,
    COMMAND_RESULT_CODE_CATEGORY_EMPTY: command_result_processor_category_empty,
    COMMAND_RESULT_CODE_CATEGORY_REQUIRES_PARAMETER: command_result_processor_category_requires_parameter,
    COMMAND_RESULT_CODE_CATEGORY_UNKNOWN_SUB_COMMAND: command_result_processor_category_unknown_sub_command,
    COMMAND_RESULT_CODE_COMMAND_REQUIRED: command_result_processor_command_required,
    COMMAND_RESULT_CODE_CALL: command_result_processor_call,
    COMMAND_RESULT_CODE_COMMAND_NOT_FOUND: command_result_processor_command_not_found,
}



__all__ = (
    'create_emoji_from_exclusive_data', 'create_partial_emoji_data', 'create_partial_emoji_from_data',
    'create_partial_emoji_from_id', 'create_partial_emoji_from_inline_data', 'create_unicode_emoji',
    'put_exclusive_emoji_data_into', 'put_partial_emoji_inline_data_into'
)

import warnings

from scarletio import export, include

from ...core import EMOJIS, UNICODE_TO_EMOJI

from .emoji import Emoji


Unicode = include('Unicode')


def _create_new_unicode(unicode_string):
    """
    Creates a new emoji from the given `unicode_string`.
    
    Parameters
    ----------
    unicode_string : `str`
        Unicode string to create emoji from.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    unicode_bytes = unicode_string.encode()
    
    warnings.warn(
        f'\nUndefined emoji : {unicode_bytes!r}\nPlease open an issue with this message.',
        RuntimeWarning,
    )
    
    return Emoji._create_unicode(Unicode('', unicode_bytes, False, None, None), False)


@export
def create_partial_emoji_from_data(data):
    """
    Creates an emoji from partial emoji data sent by Discord.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    return _create_partial_emoji_from_fields(
        data.get('name', None), data.get('id', None), data.get('animated', False),
    )


@export
def create_partial_emoji_from_inline_data(data):
    """
    Creates an emoji from inline partial emoji data sent by Discord.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    return _create_partial_emoji_from_fields(
        data.get('emoji_name', None), data.get('emoji_id', None), data.get('emoji_animated', False)
    )


def _create_partial_emoji_from_fields(emoji_name, emoji_id, emoji_animated):
    """
    Creates a partial emoji from the given fields.
    
    Parameters
    ----------
    emoji_name : `None`, `str`
        The emoji's name.
    emoji_id : `None`, `str`
        The emoji's identifier.
    emoji_animated : `bool`
        Whether the emoji is animated.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    if emoji_name is None:
        return None
    
    if emoji_id is None:
        try:
            emoji = UNICODE_TO_EMOJI[emoji_name]
        except KeyError:
            emoji = _create_new_unicode(emoji_name)
    
    else:
        # name can change
        if emoji_name is None:
            emoji_name = ''
        
        emoji = Emoji._create_partial(int(emoji_id), emoji_name, emoji_animated)
    
    return emoji


def create_partial_emoji_from_id(emoji_id):
    """
    Creates a partial emoji from the given id.
    
    Parameters
    ----------
    emoji_id : `int`
        The emoji's identifier.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        emoji = EMOJIS[emoji_id]
    except KeyError:
        emoji = Emoji._create_empty(emoji_id)
        EMOJIS[emoji_id] = emoji
    
    return emoji


@export
def create_partial_emoji_data(emoji):
    """
    Creates partial emoji data form the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The emoji to serialize.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
        The serialized emoji data.
    """
    emoji_data = {}
    unicode = emoji.unicode
    if unicode is None:
        emoji_data['id'] = str(emoji.id)
        emoji_data['name'] = emoji.name
        
        if emoji.animated:
            emoji_data['animated'] = True
    else:
        emoji_data['name'] = unicode
    
    return emoji_data


def put_partial_emoji_inline_data_into(emoji, data):
    """
    Familiar to ``create_partial_emoji_data``, but instead of creating a standalone emoji data, uses the `emoji_`
    prefix field form to add it to an already defined dictionary.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji to serialize.
    data : `dict` of (`str`, `object`) items
        The data to put the emoji fields into.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (emoji is None):
        # Require at least the `emoji_name` field.
        data['emoji_name'] = None
    
    else:
        unicode = emoji.unicode
        if unicode is None:
            data['emoji_id'] = str(emoji.id)
            data['emoji_name'] = emoji.name
            
            if emoji.animated:
                data['emoji_animated'] = True
        else:
            data['emoji_name'] = unicode
        
    return data


def create_emoji_from_exclusive_data(data):
    """
    Creates partial emoji from the given exclusive emoji data.
    
    Exclusive means, that we expect exactly `1` emoji field to be set.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : `None`, ``Emoji``
    """
    emoji_name = data.get('emoji_name', None)
    emoji_id = data.get('emoji_id', None)
    
    if (emoji_name is not None):
        try:
            emoji = UNICODE_TO_EMOJI[emoji_name]
        except KeyError:
            emoji = _create_new_unicode(emoji_name)
    
    elif (emoji_id is not None):
        emoji = create_partial_emoji_from_id(int(emoji_id))
    
    else:
        emoji = None
    
    return emoji


def put_exclusive_emoji_data_into(emoji, data):
    """
    The reversed function of ``create_emoji_from_exclusive_data``.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        The emoji to serialize.
    data : `dict` of (`str`, `object`) items
        The data to put the emoji fields into.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if (emoji is None):
        # Require at least the `emoji_name` field.
        data['emoji_name'] = None
    
    else:
        unicode = emoji.unicode
        if unicode is None:
            data['emoji_id'] = str(emoji.id)
        
        else:
            data['emoji_name'] = unicode
    
    return data


@export
def create_unicode_emoji(unicode):
    """
    Creates an emoji from the given unicode value.
    
    Parameters
    ----------
    unicode : `str`
        Unicode value.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        unicode_emoji = UNICODE_TO_EMOJI[unicode]
    except KeyError:
        raw_value = unicode.encode()
        warnings.warn(
            f'Undefined emoji : {raw_value!r}\nPlease open an issue with this message.',
            RuntimeWarning,
        )
        unicode_emoji = Emoji._create_unicode(Unicode('', raw_value, False, None, None), False)
    
    return unicode_emoji



__all__ = ('SlashCommandParameterAutoCompleter', )


from functools import partial as partial_func

from scarletio import RichAttributeErrorBaseType, export

from .....discord.client import Client
from .....discord.interaction import InteractionEvent

from ...converters import SlashCommandParameterConverter, get_application_command_parameter_auto_completer_converters

from ...exceptions import _register_exception_handler, handle_command_exception, test_exception_handler
from ...responding import process_command_coroutine
from ...utils import raw_name_to_display

from ..command_base_application_command.constants import APPLICATION_COMMAND_FUNCTION_DEEPNESS


@export
class SlashCommandParameterAutoCompleter(RichAttributeErrorBaseType):
    """
    Represents an application command parameter's auto completer.
    
    Attributes
    ----------
    _command_function : `async-callable˛
        The command's function to call.
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parameter_converters : `tuple` of ``ParameterConverter``
        Parsers to parse command parameters.
    _parent_reference : `None`, ``WeakReferer`` to (``SlashCommand``,
            ``SlashCommandFunction``, ``SlashCommandCategory``)
        The parent slash command of the auto completer, where it was registered to.
    name_pairs : `frozenset` of `tuple` (`str`, `str`)
        Raw - display parameter names, to which the converter should autocomplete.
    deepness : `int`
        How deep the auto completer was created. Deeper auto completers always overwrite higher ones.
    """
    __slots__ = (
        '_command_function', '_exception_handlers', '_parameter_converters', '_parent_reference', 'deepness',
        'name_pairs'
    )
    
    def __new__(cls, function, parameter_names, deepness, parent):
        """
        Creates a new ``SlashCommandParameterAutoCompleter`` with the given parameters.
        
        Parameters
        ----------
        function : `async-callable`
            The function to create auto completer from.
        parameter_names : `list` of `str`
            The names, which should be auto completed.
        deepness : `int`
            How deep the auto completer was created.
        parent : `None`, ``Slasher``, ``SlashCommand``, ``SlashCommandCategory``,
            ``SlashCommandFunction``
        """
        command, parameter_converters = get_application_command_parameter_auto_completer_converters(function)
        
        name_pairs = frozenset((name, raw_name_to_display(name)) for name in set(parameter_names))
        
        if parent is None:
            parent_reference = None
        else:
            parent_reference = parent._get_self_reference()
        
        self = object.__new__(cls)
        
        self._command_function = command
        self._parameter_converters = parameter_converters
        self.name_pairs = name_pairs
        self.deepness = deepness
        self._parent_reference = parent_reference
        self._exception_handlers = None
        
        return self
    
    
    def _bind_parent(self, new_parent):
        """
        Binds the parent application command function to self.
        
        If the auto completer is already bound to an other object, will return a new one.
        
        Parameter
        ---------
        new_parent : `None`, ``SlashCommandFunction``
            The parent to bind to self.
        
        Returns
        -------
        new : ``SlashCommandParameterAutoCompleter``
            The new auto completer function bound to the new parent.
        """
        parent_reference = self._parent_reference
        if (parent_reference is None):
            self_parent = None
        else:
            self_parent = parent_reference()
        
        if (new_parent is None):
            if (self_parent is None):
                new = self
                new._parent_reference = None
            
            else:
                new = self.copy()
                new._parent_reference = None
        
        else:
            if (self_parent is None):
                new = self
                new._parent_reference = new_parent._get_self_reference()
            
            else:
                if (new_parent is self_parent):
                    new = self
                
                else:
                    new = self.copy()
                    new._parent_reference = new_parent._get_self_reference()
        
        return new
    
    
    def copy(self):
        """
        Copies the parameter auto completer.
        
        Returns
        -------
        new : ``SlashCommandFunction``
        """
        new = object.__new__(type(self))
        
        # _command_function
        new._command_function = self._command_function
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _parameter_converters
        new._parameter_converters = self._parameter_converters
        
        # _parent_reference
        new._parent_reference = None
        
        # deepness
        new.deepness = self.deepness
        
        # name_pairs
        new.name_pairs = self.name_pairs
        
        return new
    

    def __repr__(self):
        """Returns the parameter auto completer's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name_pairs=')
        repr_parts.append(repr(self.name_pairs))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the parameter auto completer's hash value."""
        hash_value = 0
        
        # _command_function
        command_function = self._command_function
        try:
            command_function_hash_value = hash(command_function)
        except TypeError:
            command_function_hash_value = object.__hash__(command_function)
        hash_value ^= command_function_hash_value
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _parent_reference
        # Internal field
        
        # deepness
        hash_value ^= self.deepness
        
        # name_pairs
        hash_value ^= hash(self.name_pairs)
        
        return hash_value
    
    
    def __eq__(self, other):
        if type(self) is not type(other):
            return
        
        # _command_function
        if self._command_function != other._command_function:
            return False
        
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _parent_reference
        # Internal field
        
        # deepness
        if self.deepness != other.deepness:
            return False
        
        # name_pairs
        if self.name_pairs != other.name_pairs:
            return False
        
        return True
    
    
    async def invoke(self, client, interaction_event):
        """
        Calls the parameter auto completer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        """
        parameters = []
        
        for parameter_converter in self._parameter_converters:
            parameter = await parameter_converter(client, interaction_event, None)
            parameters.append(parameter)
        
        auto_completer_coroutine = self._command_function(*parameters)
        
        try:
            await process_command_coroutine(
                client,
                interaction_event,
                None,
                auto_completer_coroutine,
            )
        except GeneratorExit:
            raise
        
        except BaseException as err:
            exception = err
        
        else:
            return
        
        # Do not put this into the `except` branch.
        await handle_command_exception(
            self,
            client,
            interaction_event,
            exception,
        )
        return
    
    
    def _is_deeper_than(self, other):
        """
        Returns whether self is deeper than other.
        
        Parameters
        ----------
        other : ``SlashCommandParameterAutoCompleter``
        """
        self_deepness = self.deepness
        if self_deepness == APPLICATION_COMMAND_FUNCTION_DEEPNESS:
            return True
        
        other_deepness = other.deepness
        if other_deepness == APPLICATION_COMMAND_FUNCTION_DEEPNESS:
            return False
        
        if self_deepness > other_deepness:
            return True
        
        return False
    
    
    def _difference_match_parameters(self, auto_completable_parameters):
        """
        Matches auto completable parameters returning a list of the matched ones.
        
        Parameters
        ----------
        auto_completable_parameters : `set` of ``SlashCommandParameterConverter``
            Auto completable parameters.
        
        Returns
        -------
        matched : `list` of ``SlashCommandParameterConverter``
            The matched parameters.
        """
        matched = []
        
        name_pairs = set(self.name_pairs)
        
        for name_pair in list(name_pairs):
            name = name_pair[1]
            
            for parameter in auto_completable_parameters:
                if parameter.name == name:
                    name_pairs.discard(name_pair)
                    matched.append(parameter)
                    auto_completable_parameters.discard(parameter)
                    break
        
        for name_pair in list(name_pairs):
            name = name_pair[0]
            
            for parameter in auto_completable_parameters:
                if parameter.parameter_name == name:
                    name_pairs.discard(name_pair)
                    matched.append(parameter)
                    auto_completable_parameters.discard(parameter)
                    break
        
        return matched
    
    
    def error(self, exception_handler = None, *, first = False):
        """
        Registers an exception handler to the ``SlashCommandParameterAutoCompleter``.
        
        Parameters
        ----------
        exception_handler : `None`, `CoroutineFunction` = `True`, Optional
            Exception handler to register.
        first : `bool` = `False`, Optional (Keyword Only)
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler / wrapper : `CoroutineFunction` / `functools.partial`
            If `exception_handler` is not given, returns a wrapper.
        """
        if exception_handler is None:
            return partial_func(_register_exception_handler, first)
        
        return self._register_exception_handler(exception_handler, first)
    
    
    def _register_exception_handler(self, exception_handler, first):
        """
        Registers an exception handler to the ``SlashCommandParameterAutoCompleter``.
        
        Parameters
        ----------
        exception_handler : `CoroutineFunction`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler : `CoroutineFunction`
        """
        test_exception_handler(exception_handler)
        
        exception_handlers = self._exception_handlers
        if exception_handlers is None:
            self._exception_handlers = exception_handlers = []
        
        if first:
            exception_handlers.insert(0, exception_handler)
        else:
            exception_handlers.append(exception_handler)
        
        return exception_handler



__all__ = (
    'create_button', 'create_channel_select', 'create_mentionable_select', 'create_role_select', 'create_row',
    'create_string_select', 'create_text_input', 'create_user_select'
)

from .component import Component, ComponentType
from .component_metadata.constants import MAX_LENGTH_DEFAULT, MAX_VALUES_DEFAULT, MIN_LENGTH_DEFAULT, MIN_VALUES_DEFAULT


def create_button(
    label = None,
    emoji = None,
    *,
    custom_id = None,
    enabled = True,
    style = None,
    url = None,
):
    """
    Creates a new button component.
    
    Parameters
    -----------
    label : `None`, `str` = `None`, Optional
        Label of the component.
    
    emoji : `None`, ``Emoji`` = `None`, Optional
        Emoji of the button if applicable.
    
    custom_id : `None`, `str` = `None`, Optional (Keyword only)
        Custom identifier to detect which button was clicked by the user.
        
        > Mutually exclusive with the `url` field.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    style : `None`, ``ButtonStyle``, `int` = `None`, Optional (Keyword only)
        The button's style.
    
    url : `None`, `str` = `None`, Optional (Keyword only)
        Url to redirect to when clicking on the button.
        
        > Mutually exclusive with the `custom_id` field.
    
    Returns
    -------
    button : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
        - `url` and `custom_id` are mutually exclusive.
    """
    return Component(
        ComponentType.button,
        button_style = style,
        custom_id = custom_id,
        label = label,
        enabled = enabled,
        emoji = emoji,
        url = url,
    )


def create_row(
    *components,
):
    """
    Creates a row component from the given components.
    
    Parameters
    ----------
    *components : `None`, `iterable` of ``Component``
        Sub components.
    
    Returns
    -------
    row : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    """
    return Component(
        ComponentType.row,
        components = components,
    )


def create_string_select(
    options,
    custom_id = None,
    *,
    enabled = True,
    placeholder = None,
    max_values = MAX_VALUES_DEFAULT,
    min_values = MIN_VALUES_DEFAULT,
):
    """
    Creates a new string select.
    
    Parameters
    ----------
    options : `None` or `iterable` of ``StringSelectOption``
        Options of the select.
    
    custom_id : `None`, `str` = `None`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int` = `MAX_VALUES_DEFAULT`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int` = `MIN_VALUES_DEFAULT`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    string_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.string_select,
        custom_id = custom_id,
        enabled = enabled,
        options = options,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_text_input(
    label = None,
    *,
    custom_id = None,
    max_length = MAX_LENGTH_DEFAULT,
    min_length = MIN_LENGTH_DEFAULT,
    placeholder = None,
    required = None,
    style = None,
    value = None,
):
    """
    Creates a new text input.
    
    Attributes
    ----------
    label : `None`, `str` = `None`, Optional
        Label of the component.
    
    custom_id : `None`, `str` = `None`, Optional (Keyword only)
        Custom identifier to detect which text input was clicked by the user.
        
        If not given, is autogenerated form the `label` parameter.
    
    max_length : `int` = `MAX_LENGTH_DEFAULT`, Optional (Keyword only)
        The maximal length of the inputted text.
    
    min_length : `int` = `MIN_LENGTH_DEFAULT`, Optional (Keyword only)
        The minimal length of the inputted text.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    required : `None`, `bool` = `None`, Optional (Keyword only)
        Whether the field is required to be fulfilled.
        
        If not given, or given as `None`, will default to `True` if `min_length` is defined as higher than `0`.
    
    style : `None`, ``TextInputStyle``, `int` = `None`, Optional (Keyword only)
        The text input's style.
    
    value : `None`, `str` = `None`, Optional (Keyword only)
        The text input's default value.
    
    Returns
    -------
    text_input : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.text_input,
        custom_id = custom_id,
        label = label,
        max_length = max_length,
        min_length = min_length,
        placeholder = placeholder,
        required = required,
        text_input_style = style,
        value = value,
    )


def create_user_select(
    custom_id = None,
    *,
    enabled = True,
    placeholder = None,
    max_values = MAX_VALUES_DEFAULT,
    min_values = MIN_VALUES_DEFAULT,
):
    """
    Creates a new user select.
    
    Parameters
    ----------
    custom_id : `None`, `str` = `None`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int` = `MAX_VALUES_DEFAULT`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int` = `MIN_VALUES_DEFAULT`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    user_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.user_select,
        custom_id = custom_id,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_role_select(
    custom_id = None,
    *,
    enabled = True,
    placeholder = None,
    max_values = MAX_VALUES_DEFAULT,
    min_values = MIN_VALUES_DEFAULT,
):
    """
    Creates a new role select.
    
    Parameters
    ----------
    custom_id : `None`, `str` = `None`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int` = `MAX_VALUES_DEFAULT`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int` = `MIN_VALUES_DEFAULT`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    role_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.role_select,
        custom_id = custom_id,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_mentionable_select(
    custom_id = None,
    *,
    enabled = True,
    placeholder = None,
    max_values = MAX_VALUES_DEFAULT,
    min_values = MIN_VALUES_DEFAULT,
):
    """
    Creates a new mentionable (user & role) select.
    
    Parameters
    ----------
    custom_id : `None`, `str` = `None`, Optional
        Custom identifier to detect which component was used by the user.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int` = `MAX_VALUES_DEFAULT`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int` = `MIN_VALUES_DEFAULT`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    mentionable_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.mentionable_select,
        custom_id = custom_id,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )


def create_channel_select(
    custom_id = None,
    *,
    channel_types = None,
    enabled = True,
    placeholder = None,
    max_values = MAX_VALUES_DEFAULT,
    min_values = MIN_VALUES_DEFAULT,
):
    """
    Creates a new channel (user & role) select.
    
    Parameters
    ----------
    custom_id : `None`, `str` = `None`, Optional
        Custom identifier to detect which component was used by the user.
    
    channel_types : `None`, `tuple` of (``ChannelType``, `int`), Optional (Keyword only)
        The allowed channel types by the select.
    
    enabled : `bool` = `True`, Optional (Keyword only)
        Whether the button is enabled.
    
    placeholder : `None`, `str` = `None`, Optional (Keyword only)
        Placeholder text of the select.
    
    max_values : `int` = `MAX_VALUES_DEFAULT`, Optional (Keyword only)
        The maximal amount of options to select.
        
    min_values : `int` = `MIN_VALUES_DEFAULT`, Optional (Keyword only)
        The minimal amount of options to select.
    
    Returns
    -------
    channel_select : ``Component``
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
    ValueError
        - If a parameter's value is incorrect.
    """
    return Component(
        ComponentType.channel_select,
        channel_types = channel_types,
        custom_id = custom_id,
        enabled = enabled,
        max_values = max_values,
        min_values = min_values,
        placeholder = placeholder,
    )



__all__ = ()

from .constants import KEY_LENGTH_MAX


CHARACTER_RANGE = range(b'a'[0], b'z'[0] + 1)


def _convert_character(character):
    """
    Converts the given character to an allowed one if applicable.
    
    Parameters
    ----------
    character : `str`
        The character to convert.
    
    Returns
    -------
    character : `str`
    """
    character = character.casefold()
    if ord(character) in CHARACTER_RANGE:
        return character
    
    if character == '_':
        return character
    
    if character in {'-', ' '}:
        return '_'
    
    return ''


def escape_name_to_key(name):
    """
    Tries to convert the given name.
    
    Parameters
    ----------
    name : `str`
        Metadata name.
    
    Returns
    -------
    escaped : `str`
    """
    return (''.join([_convert_character(character) for character in name]))[:KEY_LENGTH_MAX]



__all__ = ()

from ...channel import Channel
from ...field_parsers import (
    bool_parser_factory, entity_id_array_parser_factory, entity_id_parser_factory, nullable_object_array_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, nullable_entity_array_putter_factory,
    optional_entity_id_array_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_array_validator_factory, entity_id_validator_factory,
    nullable_object_array_validator_factory
)
from ...guild import Guild

from ..onboarding_prompt import OnboardingPrompt

# default_channel_ids

parse_default_channel_ids = entity_id_array_parser_factory('default_channel_ids')
put_default_channel_ids_into = optional_entity_id_array_optional_putter_factory('default_channel_ids')
validate_default_channel_ids = entity_id_array_validator_factory('default_channel_ids', Channel)

# enabled

parse_enabled = bool_parser_factory('enabled', True)
put_enabled_into = bool_optional_putter_factory('enabled', True)
validate_enabled = bool_validator_factory('enabled', True)

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)

# prompts

parse_prompts = nullable_object_array_parser_factory('prompts', OnboardingPrompt)
put_prompts_into = nullable_entity_array_putter_factory('prompts', OnboardingPrompt)
validate_prompts = nullable_object_array_validator_factory('prompts', OnboardingPrompt)



__all__ = ()

import warnings

from scarletio import modulize

from ....utils.module_deprecation import deprecated_import

from .preinstanced import ChannelType


@deprecated_import
def get_channel_type_name(channel_type):
    """
    Returns the channel type's name.
    
    Parameters
    ----------
    channel_type : `int`, ``ChannelType``
        The channel type to get it's name.
    
    Returns
    -------
    channel_name : `str`
    """
    warnings.warn(
        '`get_channel_type_name` is deprecated and will be removed in 2023 January.',
        FutureWarning,
        stacklevel = 2,
    )
    
    if isinstance(channel_type, int):
        return ChannelType.get(channel_type).name
    
    if isinstance(channel_type, ChannelType):
        return channel_type.name
    
    return ''


@deprecated_import
def get_channel_type_names(channel_types):
    """
    Returns multiple channel type's names connected by comma.
    
    Parameters
    ----------
    channel_types : `iterable` of (`int`, ``ChannelType``)
        The channel types to get the their name.
    
    Returns
    -------
    channel_names : `str`
    """
    warnings.warn(
        '`get_channel_type_names` is deprecated and will be removed in 2023 January.',
        FutureWarning,
        stacklevel = 2,
    )
    
    return ', '.join(sorted(get_channel_type_name(channel_type) for channel_type in channel_types))


@deprecated_import
@modulize
class CHANNEL_TYPES:
    """
    Deprecated and will be removed in 2023 January.
    
    Contains channel type identifiers.
    
    +---------------------------------------+-------+
    | Name                                  | Value |
    +=======================================+=======+
    | guild_text                            | 0     |
    +---------------------------------------+-------+
    | private                               | 1     |
    +---------------------------------------+-------+
    | guild_voice                           | 2     |
    +---------------------------------------+-------+
    | private_group                         | 3     |
    +---------------------------------------+-------+
    | guild_category                        | 4     |
    +---------------------------------------+-------+
    | guild_announcements                   | 5     |
    +---------------------------------------+-------+
    | guild_store                           | 6     |
    +---------------------------------------+-------+
    | thread                                | 9     |
    +---------------------------------------+-------+
    | guild_thread_announcements            | 10    |
    +---------------------------------------+-------+
    | guild_thread_public                   | 11    |
    +---------------------------------------+-------+
    | guild_thread_private                  | 12    |
    +---------------------------------------+-------+
    | guild_stage                           | 13    |
    +---------------------------------------+-------+
    | guild_directory                       | 14    |
    +---------------------------------------+-------+
    | guild_forum                           | 15    |
    +---------------------------------------+-------+
    
    In addition also extra groups are defined:
    
    +---------------------------------------+-------------------------------+
    | Name                                  | Elements                      |
    +=======================================+===============================+
    | GROUP_MESSAGEABLE                     | guild_text,                   |
    |                                       | private,                      |
    |                                       | guild_voice,                  |
    |                                       | private_group,                |
    |                                       | guild_announcements,          |
    |                                       | guild_thread_announcements,   |
    |                                       | guild_thread_public,          |
    |                                       | guild_thread_private          |
    +---------------------------------------+-------------------------------+
    | GROUP_GUILD_MESSAGEABLE               | guild_text,                   |
    |                                       | guild_voice,                  |
    |                                       | guild_announcements,          |
    |                                       | guild_thread_announcements,   |
    |                                       | guild_thread_public,          |
    |                                       | guild_thread_private          |
    +---------------------------------------+-------------------------------+
    | GROUP_GUILD_MAIN_TEXT                 | guild_text,                   |
    |                                       | guild_announcements           |
    +---------------------------------------+-------------------------------+
    | GROUP_CONNECTABLE                     | private,                      |
    |                                       | guild_voice,                  |
    |                                       | private_group,                |
    |                                       | guild_stage                   |
    +---------------------------------------+-------------------------------+
    | GROUP_GUILD_CONNECTABLE               | guild_voice,                  |
    |                                       | guild_stage                   |
    +---------------------------------------+-------------------------------+
    | GROUP_PRIVATE                         | private,                      |
    |                                       | private_group                 |
    +---------------------------------------+-------------------------------+
    | GROUP_GUILD                           | guild_text,                   |
    |                                       | guild_voice,                  |
    |                                       | guild_category,               |
    |                                       | guild_announcements,          |
    |                                       | guild_store,                  |
    |                                       | guild_thread_announcements,   |
    |                                       | guild_thread_public,          |
    |                                       | guild_thread_private,         |
    |                                       | guild_stage,                  |
    |                                       | guild_directory,              |
    |                                       | guild_forum                   |
    +---------------------------------------+-------------------------------+
    | GROUP_THREAD                          | guild_thread_announcements,   |
    |                                       | guild_thread_public,          |
    |                                       | guild_thread_private          |
    +---------------------------------------+-------------------------------+
    | GROUP_CAN_CONTAIN_THREADS             | guild_text,                   |
    |                                       | guild_announcements,          |
    |                                       | guild_forum                   |
    +---------------------------------------+-------------------------------+
    | GROUP_IN_PRODUCTION                   | guild_text,                   |
    |                                       | private,                      |
    |                                       | guild_voice,                  |
    |                                       | private_group,                |
    |                                       | guild_category,               |
    |                                       | guild_announcements,          |
    |                                       | guild_thread_announcements,   |
    |                                       | guild_thread_public,          |
    |                                       | guild_thread_private,         |
    |                                       | guild_stage,                  |
    |                                       | guild_directory,              |
    |                                       | guild_forum                   |
    +---------------------------------------+-------------------------------+
    | GROUP_CAN_CREATE_INVITE_TO            | guild_text,                   |
    |                                       | guild_voice,                  |
    |                                       | private_group,                |
    |                                       | guild_announcements,          |
    |                                       | guild_store,                  |
    |                                       | guild_stage,                  |
    |                                       | guild_directory               |
    +---------------------------------------+-------------------------------+
    | GROUP_GUILD_MOVABLE                   | guild_text,                   |
    |                                       | guild_voice,                  |
    |                                       | guild_category,               |
    |                                       | guild_announcements,          |
    |                                       | guild_store,                  |
    |                                       | guild_stage,                  |
    |                                       | guild_directory,              |
    |                                       | guild_forum                   |
    +---------------------------------------+-------------------------------+
    """
    guild_text = 0
    private = 1
    guild_voice = 2
    private_group = 3
    guild_category = 4
    guild_announcements = 5
    guild_store = 6
    # 7? Not in use
    # 8? Not in use
    thread = 9 # Not in use
    guild_thread_announcements = 10
    guild_thread_public = 11
    guild_thread_private = 12
    guild_stage = 13
    guild_directory = 14
    guild_forum = 15
    
    
    GROUP_MESSAGEABLE = frozenset((
        guild_text,
        private,
        guild_voice,
        private_group,
        guild_announcements,
        guild_thread_announcements,
        guild_thread_public,
        guild_thread_private,
    ))
    
    
    GROUP_GUILD_MESSAGEABLE = frozenset((
        guild_text,
        guild_voice,
        guild_announcements,
        guild_thread_announcements,
        guild_thread_public,
        guild_thread_private,
    ))
    
    
    GROUP_GUILD_MAIN_TEXT = frozenset((
        guild_text,
        guild_announcements,
    ))
    
    
    GROUP_CONNECTABLE = frozenset((
        private,
        guild_voice,
        private_group,
        guild_stage,
    ))
    
    
    GROUP_GUILD_CONNECTABLE = frozenset((
        guild_voice,
        guild_stage,
    ))
    
    
    GROUP_PRIVATE = frozenset((
        private,
        private_group,
    ))
    
    
    GROUP_GUILD = frozenset((
        guild_text,
        guild_voice,
        guild_category,
        guild_announcements,
        guild_store,
        guild_thread_announcements,
        guild_thread_public,
        guild_thread_private,
        guild_stage,
        guild_directory,
        guild_forum,
    ))
    
    
    GROUP_THREAD = frozenset((
        guild_thread_announcements,
        guild_thread_public,
        guild_thread_private,
    ))
    
    
    GROUP_CAN_CONTAIN_THREADS = frozenset((
        guild_text,
        guild_announcements,
        guild_forum,
    ))
    
    
    GROUP_IN_PRODUCTION = frozenset((
        guild_text,
        private,
        guild_voice,
        private_group,
        guild_category,
        guild_announcements,
        guild_thread_announcements,
        guild_thread_public,
        guild_thread_private,
        guild_stage,
        guild_directory,
        guild_forum,
    ))
    
    
    GROUP_CAN_CREATE_INVITE_TO = frozenset((
        guild_text,
        guild_voice,
        private_group,
        guild_announcements,
        guild_store,
        guild_stage,
        guild_directory,
        
    ))
    
    GROUP_GUILD_MOVABLE = frozenset((
        guild_text,
        guild_voice,
        guild_category,
        guild_announcements,
        guild_store,
        guild_stage,
        guild_directory,
        guild_forum,
    ))
    
    
    CHANNEL_TYPE_NAMES = {
        guild_text: 'guild text',
        private: 'guild private',
        guild_voice: 'guild voice',
        private_group: 'private group',
        guild_category: 'guild category',
        guild_announcements: 'guild announcements',
        guild_store: 'guild store',
        thread: 'thread',
        guild_thread_announcements: 'guild thread announcements',
        guild_thread_public: 'guild thread public',
        guild_thread_private: 'guild thread private',
        guild_stage: 'guild stage',
        guild_directory: 'guild directory',
        guild_forum: 'guild forum',
    }
    
    
    DEFAULT_CHANNEL_TYPE_NAME = 'unknown'
    
    
    get_channel_type_name = get_channel_type_name
    get_channel_type_names = get_channel_type_names



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, negated_bool_parser_factory,
    nullable_date_time_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, force_bool_putter_factory, force_string_putter_factory,
    negated_bool_optional_putter_factory, nullable_date_time_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    nullable_date_time_validator_factory
)
from ..user import ClientUserBase

# channel_id

parse_channel_id = entity_id_parser_factory('channel_id')
put_channel_id_into = entity_id_putter_factory('channel_id')
validate_channel_id = entity_id_validator_factory('channel_id', NotImplemented, include = 'Channel')

# deaf

parse_deaf = bool_parser_factory('deaf', False)
put_deaf_into = force_bool_putter_factory('deaf')
validate_deaf = bool_validator_factory('deaf', False)

# guild_id

validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# mute

parse_mute = bool_parser_factory('mute', False)
put_mute_into = force_bool_putter_factory('mute')
validate_mute = bool_validator_factory('mute', False)

# requested_to_speak_at

parse_requested_to_speak_at = nullable_date_time_parser_factory('request_to_speak_timestamp')
put_requested_to_speak_at_into = nullable_date_time_optional_putter_factory('request_to_speak_timestamp')
validate_requested_to_speak_at = nullable_date_time_validator_factory('requested_to_speak_at')

# self_deaf

parse_self_deaf = bool_parser_factory('self_deaf', False)
put_self_deaf_into = force_bool_putter_factory('self_deaf')
validate_self_deaf = bool_validator_factory('self_deaf', False)

# self_mute

parse_self_mute = bool_parser_factory('self_mute', False)
put_self_mute_into = force_bool_putter_factory('self_mute')
validate_self_mute = bool_validator_factory('self_mute', False)

# self_stream

parse_self_stream = bool_parser_factory('self_stream', False)
put_self_stream_into = bool_optional_putter_factory('self_stream', False)
validate_self_stream = bool_validator_factory('self_stream', False)

# self_video

parse_self_video = bool_parser_factory('self_video', False)
put_self_video_into = force_bool_putter_factory('self_video')
validate_self_video = bool_validator_factory('self_video', False)

# session_id

parse_session_id = force_string_parser_factory('session_id')
put_session_id_into = force_string_putter_factory('session_id')
validate_session_id = force_string_validator_factory('session_id', 0, 1024)

# speaker

parse_speaker = negated_bool_parser_factory('suppress', False)
put_speaker_into = negated_bool_optional_putter_factory('suppress', False)
validate_speaker = bool_validator_factory('speaker', False)

# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id_into = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ('IntegrationAccount', )

from scarletio import RichAttributeErrorBaseType

from .fields import parse_name, parse_id, validate_name, validate_id, put_name_into, put_id_into


class IntegrationAccount(RichAttributeErrorBaseType):
    """
    Account for who an ``Integration`` is for.
    
    Attributes
    ----------
    id : `str`
        The respective account's identifier.
    name : `str`
        The respective account's name.
    """
    __slots__ = ('id', 'name', )
    
    def __new__(cls, integration_account_id = ..., name = ...):
        """
        Creates a new integration account instance from the given account data and integration type.
        
        Parameters
        ----------
        integration_account_id : `None`, `str`, Optional
            The integration account's identifier.
        name : `None`, `str`, Optional
            The integration account's name.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # integration_account_id
        if integration_account_id is ...:
            integration_account_id = ''
        else:
            integration_account_id = validate_id(integration_account_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        self = object.__new__(cls)
        self.id = integration_account_id
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new integration account instance from the given account data and integration type.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Integration account data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.name = parse_name(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the integration account into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_id_into(self.id, data, defaults)
        put_name_into(self.name, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the integration account's representation."""
        return f'<{self.__class__.__name__} id = {self.id!r}, name = {self.name!r}>'
    
    
    def __eq__(self, other):
        """Returns whether the two integration accounts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the integration account."""
        hash_value = 0
        
        # id
        hash_value ^= hash(self.id)
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty integration account, with it's attributes set as empty strings.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = ''
        self.name = ''
        return self
    
    
    def copy(self):
        """
        Copies the integration application returning a new one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = self.id
        new.name = self.name
        return new
    

    def copy_with(self, *, integration_account_id = ..., name = ...):
        """
        Copies the integration application with the given fields.
        
        Parameters
        ----------
        integration_account_id : `None`, `str`, Optional (Keyword only)
            The integration account's identifier.
        name : `None`, `str`, Optional (Keyword only)
            The integration account's name.
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # integration_account_id
        if integration_account_id is ...:
            integration_account_id = self.id
        else:
            integration_account_id = validate_id(integration_account_id)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        new = object.__new__(type(self))
        new.id = integration_account_id
        new.name = name
        return new



__all__ = ('warn_auto_reloader_availability',)

import warnings

from ..compatibility import AUTO_RELOAD_SUPPORTED, IS_LINUX


def warn_auto_reloader_availability(*, do_raise = False):
    """
    Warns if auto reload is not available.
    
    Parameters
    ----------
    do_raise : `bool` = `False`, Optional (Keyword only)
        Whether an exception should be raised instead.
    
    Raises
    ------
    RuntimeError
        - If auto reloading is no available.
    """
    if AUTO_RELOAD_SUPPORTED:
        return
    
    if not IS_LINUX:
        reason = 'Auto reloader only is available only on linux.'
    else:
        reason = 'Required package `inotify_simple` is not installed.'
    
    message = f'Plugin auto reloading is not available:\n{reason}'
    
    if do_raise:
        raise RuntimeError(message)
    
    warnings.warn(message, RuntimeWarning, stacklevel = 2)



__all__ = ('RoleManagerMetadataApplicationRoleConnection',)

from scarletio import copy_docs

from .constants import APPLICATION_ROLE_CONNECTION_KEY
from .integration import RoleManagerMetadataIntegration


class RoleManagerMetadataApplicationRoleConnection(RoleManagerMetadataIntegration):
    """
    Role manager metadata of a role managed by an application role connection.
    
    Attributes
    ----------
    integration_id : `int`
        The manager integration's identifier.
    """
    __slots__ = ()
    
    @copy_docs(RoleManagerMetadataIntegration.to_data)
    def to_data(self, *, defaults = False):
        data = RoleManagerMetadataIntegration.to_data(self)
        data[APPLICATION_ROLE_CONNECTION_KEY] = None
        return data



__all__ = ()

from ...utils import seconds_to_id_difference

from ..interaction_metadata import InteractionMetadataBase


INTERACTION_EVENT_EXPIRE_AFTER = 900 # 15 min
INTERACTION_EVENT_EXPIRE_AFTER_ID_DIFFERENCE = seconds_to_id_difference(INTERACTION_EVENT_EXPIRE_AFTER)

DEFAULT_INTERACTION_METADATA = InteractionMetadataBase()


USER_GUILD_CACHE = {}
"""
USER_GUILD_CACHE : `dict` of (`tuple` (``ClientUserBase``, ``Guild``), `int`)
    A cache which stores `user-guild` pairs as keys and their reference count as values to remember
    ``InteractionEvent``'s ``.user``-s' guild profiles of the respective ``.guild`` even if the ``Guild`` is
    uncached.

    Note, that private channel interaction, neither interactions of cached guilds are not added, what means if
    all the clients are kicked from a guild the guild profile can be lost in unexpected time.
"""



__all__ = ('PlaceHolder', 'PlaceHolderFunctional')

from scarletio import RichAttributeErrorBaseType, copy_docs, docs_property


class PlaceHolderBase(RichAttributeErrorBaseType):
    __class_doc__ = (
    """
    Base type of ``PlaceHolder`` and ``PlaceHolderFunctional``.
    
    Attributes
    ----------
    docs : `None`, `str`
        Documentation of the place held attribute.
    name : `None, `str`
        The name of the place held attribute.
    """)
    
    @property
    def __instance_doc__(self):
        return self.docs
    
    __doc__ = docs_property()
    
    __slots__ = ('docs', 'name')

    def __new__(cls, docs = None):
        """
        Creates a new slot place holder.
        
        Parameters
        ----------
        docs : `None`, `str` = `None`, Optional
            Documentation of the place held attribute.
        """
        self = object.__new__(cls)
        self.docs = docs
        self.name = None
        return self
    
    
    def __set_name__(self, owner, name):
        """
        Called when the type is constructed.
        
        Parameters
        ----------
        owner : `type`
            The parent type.
        name : `str`
            the name of the help attribute.
        """
        self.name = name
    
    
    def __get__(self, instance, type_):
        """
        Called when the place holder is accessed with a `get` operation.
        
        Returns
        -------
        self / default : ``PlaceHolder`` / `object`
            If accessed from class returns self. If from instance, returns the default object.
        """
        if instance is None:
            return self
        
        return NotImplemented
    
    
    def __set__(self, instance, value):
        """
        Called when place holder is accessed with a `set` operation.
        
        Raises
        ------
        NotImplementedError
        """
        name = self.name
        if name is None:
            name = 'unknown'
        else:
            name = repr(name)
        
        raise NotImplementedError(
            f'Setting {name} attribute of {instance.__class__} is not supported; '
            f'got instance = {instance!r}; value = {value!r}.'
        )
    
    
    def __delete__(self, instance):
        """
        Called when place holder is accessed with a `del` operation.
        
        Raises
        ------
        NotImplementedError
        """
        name = self.name
        if name is None:
            name = 'unknown'
        else:
            name = repr(name)
        
        raise NotImplementedError(
            f'Deleting {name} attribute of {instance.__class__} is not supported; '
            f'got instance = {instance!r}.'
        )


class PlaceHolder(PlaceHolderBase):
    __class_doc__ = (
    """
    Slot place holder returning a default value.
    
    Might be used to avoid `__getattr__` definitions.
    
    Attributes
    ----------
    default : `object`
        The object to return from getter.
    docs : `None`, `str`
        Documentation of the place held attribute.
    name : `None, `str`
        The name of the place held attribute.
    """)
    
    __slots__ = ('default',)
    
    
    def __new__(cls, default, docs = None):
        """
        Creates a new new slot place holder.
        
        Parameters
        ----------
        default : `object`
            The object to return from getter.
        docs : `None`, `str` = `None`, Optional
            Documentation of the place held attribute.
        """
        self = PlaceHolderBase.__new__(cls, docs)
        self.default = default
        return self
    
    
    
    @copy_docs(PlaceHolderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        return self.default


class PlaceHolderFunctional(PlaceHolderBase):
    __class_doc__ = (
    """
    Slot place holder returning a default value.
    
    Might be used to avoid `__getattr__` definitions.
    
    Attributes
    ----------
    default_function : `callable`
        A function to create the default value.
    docs : `None`, `str`
        Documentation of the place held attribute.
    name : `None, `str`
        The name of the place held attribute.
    """)
    
    __slots__ = ('default_function',)
    
    
    def __new__(cls, default_function, docs = None):
        """
        Creates a new new slot place holder.
        
        Parameters
        ----------
        default_function : `callable`
            A function to create the default value.
        docs : `None`, `str` = `None`, Optional
            Documentation of the place held attribute.
        """
        self = PlaceHolderBase.__new__(cls, docs)
        self.default_function = default_function
        return self
    
    
    @copy_docs(PlaceHolderBase.__get__)
    def __get__(self, instance, type_):
        if instance is None:
            return self
        
        return self.default_function()



__all__ = ()

from ...env import API_VERSION


if API_VERSION in (6, 7):
    PERMISSION_KEY = 'permissions_new'
    PERMISSION_ALLOW_KEY = 'allow_new'
    PERMISSION_DENY_KEY = 'deny_new'
    
else:
    PERMISSION_KEY = 'permissions'
    PERMISSION_ALLOW_KEY = 'allow'
    PERMISSION_DENY_KEY = 'deny'



__all__ = ()

from ...scheduled_event import PrivacyLevel, ScheduledEventEntityType, ScheduledEventStatus
from ...scheduled_event.scheduled_event_entity_metadata.utils import try_get_scheduled_event_metadata_type_from_data

from ..audit_log_change import AuditLogChange

from .shared import (
    _convert_preinstanced, convert_icon, convert_nothing, convert_snowflake, convert_snowflake_array,
    convert_timestamp
)


def convert_scheduled_event_entity_metadata(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        metadata_type = try_get_scheduled_event_metadata_type_from_data(before)
        if metadata_type is None:
            before = None
        else:
            before = metadata_type.from_data(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        metadata_type = try_get_scheduled_event_metadata_type_from_data(after)
        if metadata_type is None:
            after = None
        else:
            after = metadata_type.from_data(after)
    
    return AuditLogChange('entity_metadata', before, after)


def convert_scheduled_event_entity_type(name, data):
    return _convert_preinstanced('entity_type', data, ScheduledEventEntityType)


def convert_timestamp__end(name, data):
    return convert_timestamp('end', data)


def convert_timestamp__start(name, data):
    return convert_timestamp('start', data)


def convert_scheduled_event_status(name, data):
    return _convert_preinstanced('status', data, ScheduledEventStatus)


def convert_privacy_level(name, data):
    return _convert_preinstanced('privacy_level', data, PrivacyLevel)


SCHEDULED_EVENT_CONVERTERS = {
    'channel_id': convert_snowflake,
    'description': convert_nothing,
    'entity_id': convert_snowflake,
    'entity_metadata' : convert_scheduled_event_entity_metadata,
    'entity_type': convert_scheduled_event_entity_type,
    'image': convert_icon,
    'location': convert_nothing,
    'name': convert_nothing,
    'privacy_level': convert_privacy_level,
    'scheduled_end_time': convert_timestamp__end,
    'scheduled_start_time': convert_timestamp__start,
    'sku_ids': convert_snowflake_array,
    'status': convert_scheduled_event_status,
}



__all__ = ()

from ...field_parsers import entity_id_parser_factory, nullable_entity_array_parser_factory
from ...field_putters import entity_id_optional_putter_factory, nullable_entity_array_putter_factory
from ...field_validators import entity_id_validator_factory, nullable_entity_array_validator_factory

from ..soundboard_sound import SoundboardSound

# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')

# sounds

parse_sounds = nullable_entity_array_parser_factory('soundboard_sounds', SoundboardSound)
put_sounds_into = nullable_entity_array_putter_factory(
    'soundboard_sounds', SoundboardSound, force_include_internals = True
)
validate_sounds = nullable_entity_array_validator_factory('sounds', SoundboardSound)
